"""
Testing for 'copr-repo' script (Copr Backend codebase)
"""

import contextlib
import glob
import logging
import os
import runpy
import shutil
import subprocess
import tempfile
import time
from unittest import mock, skip

import distro
import munch
from packaging import version
import pytest

from testlib.repodata import load_primary_xml
from testlib import (
    assert_files_in_dir,
    AsyncCreaterepoRequestFactory,
    minimal_be_config,
)

from copr_prune_results import run_prunerepo

from copr_common.redis_helpers import get_redis_connection
from copr_backend.helpers import (
    BackendConfigReader,
    call_copr_repo,
)


modifyrepo = 'run/copr-repo'
fix_gpg_script = 'run/copr_fix_gpg.py'

# pylint: disable=attribute-defined-outside-init
# pylint: disable=too-many-public-methods

@contextlib.contextmanager
def _lock(directory="non-existent"):
    filedict = runpy.run_path(modifyrepo)
    opts = munch.Munch()
    opts.log = logging.getLogger()
    opts.directory = directory
    lock = filedict['lock']
    lockdir = os.environ.get("COPR_TESTSUITE_LOCKPATH")
    with lock(opts.directory, lockdir=lockdir, timeout=-1, log=opts.log):
        yield opts

class TestModifyRepo(object):
    def setup_method(self, method):
        self.workdir = tempfile.mkdtemp(prefix="copr-test-copr-repo")
        self.be_config = minimal_be_config(self.workdir)
        self.os_env_patcher = mock.patch.dict(os.environ, {
            'PATH': os.environ['PATH']+':run',
            'COPR_TESTSUITE_LOCKPATH': self.workdir,
            'COPR_BE_CONFIG': self.be_config,
        })
        self.os_env_patcher.start()
        self.redis = get_redis_connection(
            BackendConfigReader(self.be_config).read())
        self.request_createrepo = AsyncCreaterepoRequestFactory(self.redis)

    def teardown_method(self, method):
        shutil.rmtree(self.workdir)
        self.os_env_patcher.stop()
        self.redis.flushdb()

    def test_copr_modifyrepo_locks(self):
        with _lock(self.workdir) as opts:
            cmd = [modifyrepo, opts.directory, '--log-to-stdout']
            proc = subprocess.Popen(cmd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            try:
                proc.communicate(timeout=2)
                assert 0 # this shouldn't happen
            except subprocess.TimeoutExpired:
                proc.kill()
                out, err = proc.communicate()
                assert b"acquired lock" not in out
                assert b"acquired lock" not in err

        # with released lock
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        out, err = proc.communicate(timeout=5)
        proc.kill()
        assert b"acquired lock" in err

    @mock.patch.dict(os.environ, {'COPR_TESTSUITE_NO_OUTPUT': '1'})
    def test_copr_repo_add_subdir(self, f_second_build):
        _unused = self
        ctx = f_second_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')
        empty_repodata = load_primary_xml(repodata)

        assert empty_repodata['names'] == set()
        assert call_copr_repo(chrootdir, add=[ctx.builds[0]])
        first_repodata = load_primary_xml(repodata)

        assert first_repodata['hrefs'] == {'00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm'}
        assert call_copr_repo(chrootdir, add=[ctx.builds[1]])
        second_repodata = load_primary_xml(repodata)
        assert second_repodata['hrefs'] == {
            '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
            '00000002-example/example-1.0.4-1.fc23.x86_64.rpm'
        }
        assert_files_in_dir(chrootdir,
                            ["00000002-example", "00000001-prunerepo"], [])

    def test_copr_repo_batched_createrepo(self, f_second_build):
        ctx = f_second_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')
        assert call_copr_repo(chrootdir, add=[ctx.builds[0]])
        first_repodata = load_primary_xml(repodata)
        assert first_repodata['hrefs'] == {
            '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
        }
        # call copr-repo for second build while separate request for removal of
        # the first repo was requested
        self.request_createrepo.get(chrootdir, {
            "add": [],
            "delete": ["00000001-prunerepo"],
        })
        assert call_copr_repo(chrootdir, add=[ctx.builds[1]])

        second_repodata = load_primary_xml(repodata)
        assert second_repodata['hrefs'] == {
            '00000002-example/example-1.0.4-1.fc23.x86_64.rpm'
        }
        assert_files_in_dir(chrootdir, ["00000002-example"],
                            ["00000001-prunerepo"])

    def test_copr_repo_batched_already_processed(self, f_second_build):
        ctx = f_second_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == set()
        with _lock(chrootdir) as opts:
            # delay processing by lock
            cmd = [modifyrepo, "--batched", "--log-to-stdout",
                   opts.directory, "--add", ctx.builds[1]]
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            try:
                # start the process
                proc.communicate(timeout=2)
                assert 0 # this shouldn't happen
            except subprocess.TimeoutExpired:
                pass
            while True:
                keys = self.redis.keys("createrepo_batched*")
                if len(keys) == 1:
                    break
            key = keys[0]
            # claim we did it!
            self.redis.hset(key, "status", "success")

        (out, err) = proc.communicate()
        assert out == b""
        err_decoded = err.decode("utf-8")
        assert "Task processed by other process" in err_decoded
        assert proc.returncode == 0
        # nothing changed, copr-repo went no-op because it thinks we already
        # processed it
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == set()
        assert_files_in_dir(chrootdir,
                            ["00000001-prunerepo", "00000002-example"], [])

    def test_copr_repo_batched_two_builds(self, f_third_build):
        """ Two finished builds requesting createrepo at the same time """
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == set()
        assert call_copr_repo(chrootdir, add=[ctx.builds[1]])
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == {
            '00000002-example/example-1.0.4-1.fc23.x86_64.rpm'
        }
        # other process requested adding build[0]
        self.request_createrepo.get(chrootdir, {
            "add": [ctx.builds[0]],
            "delete": [],
        })
        assert call_copr_repo(chrootdir, add=[ctx.builds[1]])
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == {
            '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
            '00000002-example/example-1.0.4-1.fc23.x86_64.rpm'
        }

        # build 3 stays undeleted
        assert_files_in_dir(chrootdir,
                            ["00000001-prunerepo", "00000002-example",
                             "00000003-example"], [])

    def test_copr_repo_batched_full(self, f_third_build):
        """
        Full createrepo which also does one removal and one addition
        for others.
        """
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')

        # createrepo was not run for the packages, yet
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == set()

        # one other process requested removal, another addition
        self.request_createrepo.get(chrootdir, {
            "add": [],
            "delete": [ctx.builds[1]],
        })
        self.request_createrepo.get(chrootdir, {
            "add": [ctx.builds[0]],
            "delete": [],
        })

        # merged full createrepo run, still does the removal
        assert call_copr_repo(chrootdir)

        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == {
            '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
            '00000003-example/example-1.0.14-1.fc30.x86_64.rpm',
        }
        assert_files_in_dir(chrootdir,
                            ["00000001-prunerepo", "00000003-example"],
                            ["00000002-example"])

    def test_copr_repo_batched_others_full(self, f_third_build):
        """
        We add one build, but other request is to
        - remove one build
        - run full createrepo
        """
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')

        # check no one run craterepo against the builds
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == set()

        # one build finished (the second one)
        assert call_copr_repo(chrootdir, add=[ctx.builds[1]])
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == {
            '00000002-example/example-1.0.4-1.fc23.x86_64.rpm'
        }

        # other process requested full run
        self.request_createrepo.get(chrootdir, {
            "add": [],
            "delete": [],
            "full": True,
        })

        # other requested removal of first build
        self.request_createrepo.get(chrootdir, {
            "add": [],
            "delete": [ctx.builds[0]],
            "full": True,
        })

        # we request addition of third build
        assert call_copr_repo(chrootdir, add=[ctx.builds[2]])
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == {
            '00000002-example/example-1.0.4-1.fc23.x86_64.rpm',
            '00000003-example/example-1.0.14-1.fc30.x86_64.rpm',
        }
        assert_files_in_dir(chrootdir,
                            ["00000002-example", "00000003-example"],
                            ["00000001-prunerepo"])

    def test_copr_repo_add_del_mixup(self, f_third_build):
        """
        Check that if one process requests adding one build, and another
        removal, we still remove it.
        """
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')

        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == set()

        # delete request
        self.request_createrepo.get(chrootdir, {
            "add": [],
            "delete": [ctx.builds[1]],
        })

        # and add request for the same build
        assert call_copr_repo(chrootdir, add=[ctx.builds[1]])

        repoinfo = load_primary_xml(repodata)
        output = subprocess.check_output(["createrepo_c", "--version"],
                                         universal_newlines=True)
        expected_hrefs = set()
        if version.parse(output.split()[1]) < version.parse("0.16.1"):
            # https://github.com/rpm-software-management/createrepo_c/issues/222
            expected_hrefs = set([
                '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
                '00000003-example/example-1.0.14-1.fc30.x86_64.rpm',
            ])
        assert repoinfo['hrefs'] == expected_hrefs


    @mock.patch.dict(os.environ, {'COPR_TESTSUITE_NO_OUTPUT': '1'})
    def test_copr_repo_add_subdir_devel(self, f_acr_on_and_first_build):
        _unused = self
        ctx = f_acr_on_and_first_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')
        devel_repodata = os.path.join(chrootdir, 'devel', 'repodata')
        empty_repodata = load_primary_xml(repodata)
        assert empty_repodata == load_primary_xml(devel_repodata)
        assert call_copr_repo(chrootdir, add=[ctx.builds[0]], devel=True)

        # shouldn't change
        assert empty_repodata == load_primary_xml(repodata)
        updated = load_primary_xml(devel_repodata)
        assert updated['hrefs'] == {
            '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
        }

        # the --devel repdata need to have 'xml:base' element, otherwise those
        # wouldn't be able to reference ../ locations
        assert updated['packages']['prunerepo']['xml:base'] == \
                'https://example.com/results/john/empty/fedora-rawhide-x86_64'

    @pytest.mark.parametrize('add', [
        ["aaa", ''],
        ["File 1"],
        ["slash/in/path"],
        [".."],
    ])
    def test_copr_repo_subdir_validator(self, add, caplog):
        log = logging.getLogger()
        assert call_copr_repo('/some/dir', add=add, logger=log) == 0
        # not using capsys because pytest/issues/5997
        messages = [r.message for r in caplog.records]
        assert any(['copr-repo: error: argument' in m for m in messages])

    @mock.patch("copr_backend.helpers.subprocess.Popen")
    def test_copr_repo_subdir_none_doesnt_raise(self, popen):
        """ check that None is skipped in add (or delete) """
        popen.return_value.communicate.return_value = ("", "")
        popen.return_value.returncode = 0
        assert True == call_copr_repo('/some/dir', add=['xxx', None])
        assert len(popen.call_args_list) == 1
        call = popen.call_args_list[0]
        assert call[0][0] == ['copr-repo', '--batched', '/some/dir', '--add', 'xxx']

    @mock.patch("copr_backend.helpers.subprocess.Popen")
    def test_copr_repo_do_stat(self, popen):
        """ do_stat=True adds --do-stat option """
        popen.return_value.communicate.return_value = ("", "")
        popen.return_value.returncode = 0
        assert call_copr_repo("/some/dir", do_stat=True) is True
        assert len(popen.call_args_list) == 1
        call = popen.call_args_list[0]
        assert call[0][0] == ["copr-repo", "--batched", "/some/dir", "--do-stat"]

    @pytest.mark.parametrize(
        "do_stat, chroot, database_option",
        [
            pytest.param(True, "fedora-37", "--no-database"),
            pytest.param(False, "centos-stream-8", "--no-database"),
            pytest.param(True, "rhel-8", "--no-database"),
            pytest.param(False, "epel-8", "--no-database"),
        ]
    )
    @mock.patch("copr_backend.helpers.subprocess.Popen")
    def test_copr_repo_run_createrepo(self, popen, do_stat, chroot, database_option):
        """ Check that we run createrepo_c with appropriate arguments """

        # Mock the run_prunerepo properly
        popen.return_value.communicate.return_value = ("","")
        popen.return_value.returncode = 0
        repodir = os.path.join(self.workdir, "testrepo")
        repodata = os.path.join(repodir, "repodata")
        xml = os.path.join(repodata, "repomd.xml")
        os.makedirs(repodata)
        with open(xml, 'w'):
            pass
        filedict = runpy.run_path(modifyrepo)
        opts = munch.Munch()
        opts.directory = repodir
        opts.add = []
        opts.delete = []
        opts.full = True
        opts.devel = False
        opts.do_stat = do_stat
        opts.log = logging.getLogger()
        opts.rpms_to_remove = []
        opts.chroot = chroot

        # run the method
        filedict["run_createrepo"](opts)

        additional_args = [] if do_stat else ["--skip-stat"]

        assert popen.call_args_list[0][0][0] == \
            ["/usr/bin/createrepo_c", repodir, database_option,
             "--ignore-lock", "--local-sqlite", "--cachedir", "/tmp/",
             "--workers", "8", "--general-compress-type=gz", "--update"] + additional_args

    @pytest.mark.skipif(
        (distro.id() == 'fedora' and int(distro.version()) >= 36) or
        (distro.id() == 'openeuler' and float(distro.version()) >= 24.03),
        reason="createrepo_c dropped md5 checksum support"
    )
    def test_copr_repo_el5(self, f_third_build):
        """
        Test that special createrepo_c arguments are used when creating
        el5 repositories.
        """
        _unused = self
        ctx = f_third_build
        chroot = ctx.chroots[0]
        old_chrootdir = os.path.join(ctx.empty_dir, chroot)
        # assure that it looks like el5 directory
        chrootdir = os.path.join(ctx.empty_dir, "rhel-5-x86_64")
        repodata = os.path.join(chrootdir, 'repodata')
        subprocess.check_call(["cp", "-r", old_chrootdir, chrootdir])
        assert call_copr_repo(old_chrootdir, add=[ctx.builds[0]],
                              delete=[ctx.builds[2]])
        assert call_copr_repo(chrootdir, add=[ctx.builds[0]],
                              delete=[ctx.builds[2]])
        repoinfo = load_primary_xml(repodata)
        assert repoinfo['hrefs'] == {
            '00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm',
        }

        # rhel-5 contains md5 checksums
        assert_files_in_dir(chrootdir,
                            ["00000002-example", "00000001-prunerepo"],
                            ["00000003-example"])
        assert repoinfo["packages"]["prunerepo"]["chksum_type"] == "md5"

        # other chroots are sha256
        repodata = os.path.join(old_chrootdir, 'repodata')
        repoinfo = load_primary_xml(repodata)
        assert repoinfo["packages"]["prunerepo"]["chksum_type"] == "sha256"

    def test_copr_repo_noop(self, f_second_build):
        """
        When anyone requests removal (or addition) of directories which do not
        exist, there's no point in running the createrepo_c at all.
        """
        ctx = f_second_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        self.request_createrepo.get(chrootdir, {
            "add": [],
            "delete": ["non-existing-dir"],
        })
        assert call_copr_repo(chrootdir, add=["non-existing-dir-2"])
        repodata = os.path.join(chrootdir, 'repodata')
        repoinfo = load_primary_xml(repodata)
        assert repoinfo["hrefs"] == set()
        keys = self.redis.keys("createrepo_batch*")
        assert len(keys) == 1
        task_dict = self.redis.hgetall(keys[0])
        assert task_dict["status"] == "success"

    @mock.patch("copr_backend.helpers.subprocess.Popen")
    def test_copr_repo_rpms_to_remove_in_call(self, popen):
        """ check that list of rpm files to be removed is added to copr-repo call """
        popen.return_value.communicate.return_value = ("","")
        popen.return_value.returncode = 0

        assert call_copr_repo('/some/dir', rpms_to_remove=['xxx.rpm'])
        assert len(popen.call_args_list) == 1
        call = popen.call_args_list[0]
        assert call[0][0] == ['copr-repo', '--batched', '/some/dir', '--rpms-to-remove', 'xxx.rpm']

    def test_copr_repo_rpms_to_remove_passes(self, f_third_build):
        _unused = self
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)

        assert_files_in_dir(chrootdir, ["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"], [])
        assert call_copr_repo(chrootdir, rpms_to_remove=["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"])
        assert_files_in_dir(chrootdir, [], ["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"])

    def test_copr_repo_rpms_to_remove_passes_2(self, f_third_build):
        _unused = self
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)

        assert_files_in_dir(chrootdir, ["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"], [])
        assert call_copr_repo(chrootdir, rpms_to_remove=[])
        assert_files_in_dir(chrootdir, ["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"], [])

    def test_copr_repo_rpms_to_remove_passes_3(self, f_third_build):
        _unused = self
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)

        assert call_copr_repo(chrootdir, add=[ctx.builds[0]])
        assert_files_in_dir(chrootdir, ["00000001-prunerepo", "00000002-example"], [])
        assert call_copr_repo(chrootdir,
                              rpms_to_remove=["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"])
        assert_files_in_dir(chrootdir,
                            [],
                            ["00000002-example/example-1.0.4-1.fc23.x86_64.rpm"])
        assert_files_in_dir(chrootdir,
                            ["00000001-prunerepo/prunerepo-1.1-1.fc23.noarch.rpm"],
                            [])

    def test_comps_present(self, f_third_build):
        _unused = self
        ctx = f_third_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        copms_path = os.path.join(chrootdir, "comps.xml")
        with open(copms_path, "w"):
            pass

        assert call_copr_repo(chrootdir)

        name = glob.glob(os.path.join(chrootdir, "repodata", "*-comps.xml.*"))
        assert os.path.exists(name[0])
        assert name[0].endswith((".zst", ".gz"))

    @skip("Works locally and in Copr but fails in Koji")
    @mock.patch("copr_prune_results.LOG", logging.getLogger())
    def test_run_prunerepo(self, f_builds_to_prune):
        _unused = self
        ctx = f_builds_to_prune
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        assert_files_in_dir(os.path.join(chrootdir, '00999999-dummy-pkg'),
                            ["dummy-pkg-1-1.fc34.x86_64.rpm"],
                            [])
        run_prunerepo(chrootdir, 'john', 'empty', chroot, 0)
        assert_files_in_dir(os.path.join(chrootdir, '00999999-dummy-pkg'),
                            ["prune.log"],
                            ["dummy-pkg-1-1.fc34.x86_64.rpm"])
        logfile = os.path.join(chrootdir, "00999999-dummy-pkg", "prune.log")
        with open(logfile, "r") as fd:
            lines = fd.readlines()
            assert len(lines) == 1
            assert "pruned on" in lines[0]

    @mock.patch("copr_prune_results.get_rpms_to_remove")
    @mock.patch("copr_prune_results.LOG", logging.getLogger())
    def test_pruner_arg_max(self, patched_get_rpms):
        rpm_count = 10000
        rpm_name = "x"*77+".rpm"
        rpms_found = [rpm_name] * rpm_count
        patched_get_rpms.return_value = rpms_found

        with mock.patch("copr_prune_results.subprocess.Popen") as patched_popen:
            patched_popen.return_value.communicate.return_value = ("","")
            patched_popen.return_value.returncode = 0
            run_prunerepo("/tmp", "blah", "blah", "blah", 0, False)
            call = patched_popen.call_args_list[1][0][0]
            assert len(patched_popen.call_args_list) == 3

        assert list == type(call)
        assert 1000 < len(call) < 9000*2+100
        # Test we can execute.
        subprocess.check_call(["/bin/true"] + call)

        opts = ["--rpms-to-remove", rpm_name] * 9000
        assert call == ["copr-repo", "--batched", "/tmp"] + opts + ["--no-appstream-metadata"]

    @pytest.mark.parametrize('run_bg', [True, False])
    @mock.patch.dict(os.environ, {'COPR_TESTSUITE_NO_OUTPUT': '1'})
    def test_copr_repo_timeouted_check(self, f_second_build, run_bg):
        _unused = self
        ctx = f_second_build
        chroot = ctx.chroots[0]
        chrootdir = os.path.join(ctx.empty_dir, chroot)
        repodata = os.path.join(chrootdir, 'repodata')

        # empty repodata at the beginning
        empty_repodata = load_primary_xml(repodata)
        assert empty_repodata['names'] == set()

        pid = os.fork()
        if not pid:
            # give parent some time to lock the repo
            time.sleep(1)
            # Run the blocked (by parent) createrepo, it must finish soon
            # anway because parent will claim the task is done.
            assert call_copr_repo(chrootdir, add=[ctx.builds[1]])
            # sys.exit() can not be used in testsuite
            os._exit(0)  # pylint: disable=protected-access

        with _lock(chrootdir):
            # give the child some time to fill its Redis keys
            sleeper = 1
            while True:
                if len(self.redis.keys()) > 0:
                    break
                sleeper += 1
                time.sleep(0.1)
                assert sleeper < 10*15  # 15s

            assert len(self.redis.keys()) == 1
            key = self.redis.keys()[0]

            # Claim we did that task (even if not) and check that the child
            # finishes after some time.
            if not run_bg:
                self.redis.hset(key, "status", "success")
                assert os.wait()[1] == 0
                assert self.redis.get(key) is None

        if run_bg:
            # actually process the background job!
            assert os.wait()[1] == 0

        repodata = load_primary_xml(repodata)
        if run_bg:
            assert repodata['names'] == {'example'}
        else:
            assert repodata['names'] == set()


@mock.patch("copr_backend.helpers.subprocess.Popen")
def test_aws_cdn_refresh(popen):
    """ Check that we run createrepo_c with appropriate arguments """

    # Mock the run_prunerepo properly
    popen.return_value.communicate.return_value = ("","")
    popen.return_value.returncode = 0
    method = runpy.run_path(fix_gpg_script)["invalidate_aws_cloudfront_data"]

    opts = munch.Munch()
    opts.results_baseurl = "https://example.com/results"
    opts.aws_cloudfront_distribution = "XYZ"
    method(opts, "jdoe", "foo", "fedora-rawhide-x86_64")

    assert popen.call_args_list[0][0][0] == \
        ["aws", "cloudfront", "create-invalidation", "--distribution-id", "XYZ",
         "--paths", "/results/jdoe/foo/fedora-rawhide-x86_64/*"]



# coding: utf-8

import json

import logging
import logging.handlers
import os
from setproctitle import setproctitle

from copr_common.redis_helpers import get_redis_connection

# TODO: remove when RedisLogHandler works fine
from .. import constants
from .. import helpers


class RedisLogHandler(object):
    """
    Single point to collect logs through redis pub/sub and write
        them through standard python logging lib
    """

    def __init__(self, opts):
        self.opts = opts

        self.log_dir = os.path.dirname(self.opts.log_dir)
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir, mode=0o750)
        self.components = helpers.LOG_COMPONENTS

    def setup_logging(self):

        self.main_logger = logging.Logger("logger", level=logging.DEBUG)
        self.main_handler = logging.handlers.WatchedFileHandler(
            filename=os.path.join(self.log_dir, "logger.log"))
        self.main_handler.setFormatter(self.opts.log_format)
        self.main_logger.addHandler(self.main_handler)

        level = getattr(logging, self.opts.log_level.upper(), None)
        self.loggers = {}

        for component in self.components:
            logger = logging.Logger(component)
            handler = logging.handlers.WatchedFileHandler(
                filename=os.path.join(self.log_dir, "{}.log".format(component)))
            handler.setFormatter(self.opts.log_format)
            handler.setLevel(level)
            logger.addHandler(handler)
            self.loggers[component] = logger

    def handle_msg(self, json_event):
        try:
            event = json.loads(json_event)
            who = event.get('who', None)
            if not who:
                raise Exception("No LogRecord.who field, raw: {}".format(event))
            if who not in self.loggers:
                raise Exception("Unknown LogRecord.who field: {}, raw event: {}"
                                .format(who, event))

            log_record = logging.makeLogRecord(event)
            self.loggers[who].handle(log_record)

        except Exception as err:
            self.main_logger.exception(err)

    def run(self):
        self.setup_logging()
        setproctitle("RedisLogHandler")

        rc = get_redis_connection(self.opts)
        while True:
            # indefinitely wait for the next entry, note that blpop returns
            # tuple (FIFO_NAME, ELEMENT)
            (_, json_event) = rc.blpop([constants.LOG_REDIS_FIFO])
            self.handle_msg(json_event)



import os
import tempfile
import shutil
import time
from unittest import mock
from unittest.mock import MagicMock

from munch import Munch
import pytest

from copr_backend.exceptions import CoprSignError, CoprSignNoKeyError, CoprKeygenRequestError
from copr_backend.sign import (
    get_pubkey, _sign_one, sign_rpms_in_dir, create_user_keys,
    gpg_hashtype_for_chroot,
    call_sign_bin,
)

STDOUT = "stdout"
STDERR = "stderr"


class TestSign(object):
    # pylint: disable=too-many-public-methods

    def setup_method(self, method):
        self.username = "foo"
        self.projectname = "bar"

        self.usermail = "foo#bar@copr.fedorahosted.org"
        self.test_time = time.time()
        self.tmp_dir_path = None

        self.opts = Munch(keygen_host="example.com")
        self.opts.gently_gpg_sha256 = False
        self.opts.sign_domain = "fedorahosted.org"

    def teardown_method(self, method):
        if self.tmp_dir_path:
            shutil.rmtree(self.tmp_dir_path)

    @pytest.fixture
    def tmp_dir(self):
        subdir = "test_createrepo_{}".format(time.time())
        self.tmp_dir_path = os.path.join(tempfile.gettempdir(), subdir)
        os.mkdir(self.tmp_dir_path)

    @pytest.fixture
    def tmp_files(self):
        # ! require tmp_dir created before
        self.file_names = ["foo.rpm", "bar.rpm", "bad", "morebadrpm"]
        for name in self.file_names:
            path = os.path.join(self.tmp_dir_path, name)
            with open(path, "w") as handle:
                handle.write("1")

    @mock.patch("copr_backend.sign.Popen")
    def test_get_pubkey(self, mc_popen):
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, STDERR)
        mc_handle.returncode = 0
        mc_popen.return_value = mc_handle

        result = get_pubkey(self.username, self.projectname, MagicMock(), self.opts.sign_domain)
        assert result == STDOUT
        assert mc_popen.call_args[0][0] == ['/bin/sign', '-u', self.usermail, '-p']


    @mock.patch("copr_backend.sign.Popen")
    def test_get_pubkey_error(self, mc_popen):
        mc_popen.side_effect = IOError(STDERR)

        with pytest.raises(CoprSignError):
            get_pubkey(self.username, self.projectname, MagicMock(), self.opts.sign_domain)


    @mock.patch("copr_backend.sign.time.sleep")
    @mock.patch("copr_backend.sign.Popen")
    def test_get_pubkey_unknown_key(self, mc_popen, _sleep):
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, "unknown key: foobar")
        mc_handle.returncode = 1
        mc_popen.return_value = mc_handle

        with pytest.raises(CoprSignNoKeyError) as err:
            get_pubkey(self.username, self.projectname, MagicMock(), self.opts.sign_domain)

        assert "There are no gpg keys for user foo in keyring" in str(err)

    @mock.patch("copr_backend.sign.time.sleep")
    @mock.patch("copr_backend.sign.Popen")
    def test_get_pubkey_unknown_error(self, mc_popen, _sleep):
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, STDERR)
        mc_handle.returncode = 1
        mc_popen.return_value = mc_handle

        with pytest.raises(CoprSignError) as err:
            get_pubkey(self.username, self.projectname, MagicMock(), self.opts.sign_domain)

        assert "Failed to get user pubkey" in str(err)

    @mock.patch("copr_backend.sign.Popen")
    def test_get_pubkey_outfile(self, mc_popen, tmp_dir):
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, STDERR)
        mc_handle.returncode = 0
        mc_popen.return_value = mc_handle

        outfile_path = os.path.join(self.tmp_dir_path, "out.pub")
        assert not os.path.exists(outfile_path)
        result = get_pubkey(self.username, self.projectname, MagicMock(),
                            self.opts.sign_domain, outfile_path)
        assert result == STDOUT
        assert os.path.exists(outfile_path)
        with open(outfile_path) as handle:
            content = handle.read()
            assert STDOUT == content

    @mock.patch("copr_backend.sign.Popen")
    def test_sign_one(self, mc_popen):
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, STDERR)
        mc_handle.returncode = 0
        mc_popen.return_value = mc_handle

        fake_path = "/tmp/pkg.rpm"
        result = _sign_one(fake_path, self.usermail, "sha1", MagicMock())
        assert STDOUT, STDERR == result

        expected_cmd = ['/bin/sign', "-4", "-h", "sha1", "-u", self.usermail,
                        "-r", fake_path]
        assert mc_popen.call_args[0][0] == expected_cmd

    @mock.patch("copr_backend.sign.Popen")
    def test_sign_one_popen_error(self, mc_popen):
        mc_popen.side_effect = IOError()

        fake_path = "/tmp/pkg.rpm"
        with pytest.raises(CoprSignError):
            _sign_one(fake_path, self.usermail, "sha256", MagicMock())

    @mock.patch("copr_backend.sign.time.sleep")
    @mock.patch("copr_backend.sign.Popen")
    def test_sign_one_cmd_erro(self, mc_popen, _sleep):
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, STDERR)
        mc_handle.returncode = 1
        mc_popen.return_value = mc_handle

        fake_path = "/tmp/pkg.rpm"
        with pytest.raises(CoprSignError):
            _sign_one(fake_path, self.usermail, "sha256", MagicMock())

    @mock.patch("copr_backend.sign.time.sleep")
    @mock.patch("copr_backend.sign.Popen")
    def test_call_sign_bin_repeatedly(self, mc_popen, _sleep):
        """
        Test that we attempt to run /bin/sign multiple times if it returns
        non-zero exit status
        """
        mc_handle = MagicMock()
        mc_handle.communicate.return_value = (STDOUT, STDERR)
        mc_handle.returncode = 1
        mc_popen.return_value = mc_handle
        call_sign_bin(cmd=[], log=MagicMock())
        assert mc_popen.call_count == 3

    @mock.patch("copr_backend.sign.SafeRequest.send")
    def test_create_user_keys(self, mc_request):
        mc_request.return_value.status_code = 200
        create_user_keys(self.username, self.projectname, self.opts)

        assert mc_request.called
        expected_call = mock.call(
            url="http://example.com/gen_key",
            data={"name_real": "foo_bar", "name_email": "foo#bar@copr.fedorahosted.org"},
            method="post"
        )
        assert mc_request.call_args == expected_call

    @mock.patch("copr_backend.sign.SafeRequest.send")
    def test_create_user_keys_error_1(self, mc_request):
        mc_request.side_effect = IOError()
        with pytest.raises(CoprKeygenRequestError) as err:
            create_user_keys(self.username, self.projectname, self.opts)

        assert "Failed to create key-pair" in str(err)


    @mock.patch("copr_backend.sign.SafeRequest.send")
    def test_create_user_keys_err(self, mc_request):
        for code in [400, 401, 404, 500, 599]:
            mc_request.return_value.status_code = code
            mc_request.return_value.content = "error: {}".format(code)

            with pytest.raises(CoprKeygenRequestError) as err:
                create_user_keys(self.username, self.projectname, self.opts)
            assert "Failed to create key-pair for user: foo, project:bar" in str(err)

    @mock.patch("copr_backend.sign._sign_one")
    @mock.patch("copr_backend.sign.create_user_keys")
    @mock.patch("copr_backend.sign.get_pubkey")
    def test_sign_rpms_id_dir_nothing(self, mc_gp, mc_cuk, mc_so,
                                      tmp_dir):
        # empty target dir doesn't produce error
        sign_rpms_in_dir(self.username, self.projectname,
                         self.tmp_dir_path, "epel-8-x86_64", self.opts,
                         log=MagicMock())

        assert not mc_gp.called
        assert not mc_cuk.called
        assert not mc_so.called

    @mock.patch("copr_backend.sign._sign_one")
    @mock.patch("copr_backend.sign.create_user_keys")
    @mock.patch("copr_backend.sign.get_pubkey")
    def test_sign_rpms_id_dir_ok(self, mc_gp, mc_cuk, mc_so,
                                      tmp_dir, tmp_files):

        sign_rpms_in_dir(self.username, self.projectname,
                         self.tmp_dir_path, "fedora-rawhide-x86_64",
                         self.opts, log=MagicMock())

        assert mc_gp.called
        assert not mc_cuk.called
        assert mc_so.called

        pathes = [call[0][0] for call in mc_so.call_args_list]
        count = 0
        for name in self.file_names:
            if name.endswith(".rpm"):
                count += 1
                assert os.path.join(self.tmp_dir_path, name) in pathes
        assert len(pathes) == count

    @mock.patch("copr_backend.sign._sign_one")
    @mock.patch("copr_backend.sign.create_user_keys")
    @mock.patch("copr_backend.sign.get_pubkey")
    def test_sign_rpms_id_dir_error_on_pubkey(
            self, mc_gp, mc_cuk, mc_so, tmp_dir, tmp_files):

        mc_gp.side_effect = CoprSignError("foobar")
        with pytest.raises(CoprSignError):
            sign_rpms_in_dir(self.username, self.projectname,
                             self.tmp_dir_path, "epel-7-x86_64", self.opts,
                             log=MagicMock())

        assert mc_gp.called
        assert not mc_cuk.called
        assert not mc_so.called

    @mock.patch("copr_backend.sign._sign_one")
    @mock.patch("copr_backend.sign.create_user_keys")
    @mock.patch("copr_backend.sign.get_pubkey")
    def test_sign_rpms_id_dir_no_pub_key(
            self, mc_gp, mc_cuk, mc_so, tmp_dir, tmp_files):

        mc_gp.side_effect = CoprSignNoKeyError("foobar")

        sign_rpms_in_dir(self.username, self.projectname,
                         self.tmp_dir_path, "rhel-7-x86_64", self.opts,
                         log=MagicMock())

        assert mc_gp.called
        assert mc_cuk.called
        assert mc_so.called

    @mock.patch("copr_backend.sign._sign_one")
    @mock.patch("copr_backend.sign.create_user_keys")
    @mock.patch("copr_backend.sign.get_pubkey")
    def test_sign_rpms_id_dir_sign_error_one(
            self, mc_gp, mc_cuk, mc_so, tmp_dir, tmp_files):

        mc_so.side_effect = [
            None, CoprSignError("foobar"), None
        ]
        with pytest.raises(CoprSignError):
            sign_rpms_in_dir(self.username, self.projectname,
                             self.tmp_dir_path, "fedora-36-x86_64", self.opts,
                             log=MagicMock())

        assert mc_gp.called
        assert not mc_cuk.called

        assert mc_so.called

    @mock.patch("copr_backend.sign._sign_one")
    @mock.patch("copr_backend.sign.create_user_keys")
    @mock.patch("copr_backend.sign.get_pubkey")
    def test_sign_rpms_id_dir_sign_error_all(
            self, mc_gp, mc_cuk, mc_so, tmp_dir, tmp_files):

        mc_so.side_effect = CoprSignError("foobar")
        with pytest.raises(CoprSignError):
            sign_rpms_in_dir(self.username, self.projectname,
                             self.tmp_dir_path, "fedora-36-i386", self.opts,
                             log=MagicMock())

        assert mc_gp.called
        assert not mc_cuk.called

        assert mc_so.called


def test_chroot_gpg_hashes():
    chroots = [
        ("fedora-26-x86_64", "sha1"),
        ("fedora-27-s390x", "sha256"),
        ("fedora-eln-x86_64", "sha256"),
        ("fedora-rawhide-x86_64", "sha256"),
        ("mageia-8-x86_64", "sha256"),
        ("opensuse-tumbleweed-aarch64", "sha256"),
        ("epel-7-ppc64", "sha256"),
        ("centos-7.dev-aarch64", "sha256"),
        ("epel-8-aarch64", "sha256"),
        ("rhel-8.dev-ppc64le", "sha256"),
        ("oraclelinux-9-s390x", "sha256"),
        ("centos-stream-8-s390x", "sha256"),
        ("centos-stream-9-s390x", "sha256"),
        ("rhel-rawhide-s390x", "sha256"),
        # we don't expect stream 7 will ever exist
        ("centos-stream-7-aarch64", "sha256"),
        ("srpm-builds", "sha256"),
        ("epel-4-s390x", "sha1"),
        ("epel-5-s390x", "sha256"),
        ("centos-stream+epel-next-9-aarch64", "sha256"),
    ]

    opts = Munch()
    opts.gently_gpg_sha256 = False

    for chroot, exp_type in chroots:
        assert (chroot, exp_type) == (chroot, gpg_hashtype_for_chroot(chroot, opts))

    opts.gently_gpg_sha256 = True
    chroots = [
        ("fedora-26-x86_64", "sha1"),
        ("fedora-27-s390x", "sha1"),
        ("fedora-eln-x86_64", "sha1"),
        ("fedora-rawhide-x86_64", "sha1"),
        ("mageia-8-x86_64", "sha1"),
        ("opensuse-tumbleweed-aarch64", "sha1"),
        ("epel-7-ppc64", "sha1"),
        ("centos-7.dev-aarch64", "sha1"),
        ("epel-8-aarch64", "sha256"),
        ("rhel-8.dev-ppc64le", "sha256"),
        ("oraclelinux-9-s390x", "sha256"),
        ("centos-stream-8-s390x", "sha256"),
        ("centos-stream-9-s390x", "sha256"),
        ("rhel-rawhide-s390x", "sha256"),
        ("srpm-builds", "sha1"),
    ]
    for chroot, exp_type in chroots:
        assert (chroot, exp_type) == (chroot, gpg_hashtype_for_chroot(chroot, opts))



import contextlib
import logging
import os
import shlex
import subprocess
import time

import netaddr

from copr_common.subprocess_live_output import PosixPipedProcess

DEFAULT_SUBPROCESS_TIMEOUT = 180

class SSHConnectionError(Exception):
    pass


def _user_readable_command(cmd):
    if isinstance(cmd, list):
        return ' '.join([shlex.quote(x) for x in cmd])
    return cmd


class SSHConnection:
    """
    SSH connection representation.

    This class heavily depedns on the native openssh client configuration file
    (man 5 sh_config).  The example configuration might be:

        $ cat /home/copr/.ssh/builder_config
        Host *
        # For dynamically started VMs.
        StrictHostKeyChecking no
        UserKnownHostsFile /dev/null

        # For non-default paths to identity file.
        IdentityFile ~/.ssh/id_rsa_copr

        # Ensure remote command uses proper line buffering for live logs
        # (so called live logs).
        RequestTTY=force

        # Keep control sockets open, to speedup subsequent command runs.
        ControlPath=/home/copr/ssh_socket_%h_%p_%r
        ControlMaster=auto
        ControlPersist=900

    Then just use:
    SSHConnection(user='mockbuilder', host=vm_ip,
                  config_file='/home/copr/.ssh/builder_config')

    :param user:
        Remote user name.  'root' by default.
    :param host:
        Remote hostname or IP.  'localhost' by default.
    :param  config_file:
        Full (absolute) path ssh config file to be used.  None by default means
        the default ssh configuration is used /etc/ssh_config and ~/.ssh/config.
    """

    def __init__(self, user=None, host=None, config_file=None, port=None, log=None):
        # TODO: Some of the calling code places heavily re-try the ssh
        # connection..  There's a some small chance that the host goes down, and
        # some other host is started with the same hostname (or IP address).
        # Therefore we should remember the host's SSH fingerprint here and check
        # it when reconnecting.
        self.config_file = config_file
        self.user = user or 'root'
        self.host = host or 'localhost'
        self.port = port
        if log:
            self.log = log
        else:
            self.log = logging.getLogger()

    def _ssh_base(self):
        cmd = ['ssh']
        if self.config_file:
            cmd = cmd + ['-F', self.config_file]
        if self.port is not None:
            cmd += ["-p", str(self.port)]
        cmd.append('{0}@{1}'.format(self.user, self.host))
        return cmd

    @contextlib.contextmanager
    def _popen_timeouted(self, command, *args, **kwargs):
        """
        Wrap subprocess.Popen, catch TimeoutExpired errors and re-raise as
        SSHConnectionError.
        """
        timeouted = False
        with subprocess.Popen(command, *args, **kwargs) as proc:
            cmd_user = _user_readable_command(command)
            try:
                self.log.info("Popen command started: %s", cmd_user)
                yield proc
            except subprocess.TimeoutExpired:
                # We can not raise SSHConnectionError here, that would be
                # unexpected for the subprocess context manager logic!
                timeouted = True
                proc.kill()
        if timeouted:
            raise SSHConnectionError(f"Command over SSH timeouted: {cmd_user}")

    def _run(self, user_command, stdout, stderr, subprocess_timeout):
        real_command = self._ssh_base() + [user_command]
        with self._popen_timeouted(real_command, stdout=stdout, stderr=stderr,
                                   encoding="utf-8") as proc:
            retval = proc.wait(timeout=subprocess_timeout)

        if retval == 255:
            # Because we don't manage the control path (that's done in ssh
            # configuration), we can not really check that 255 exit status is
            # ssh connection error, or command error.
            raise SSHConnectionError("Connection broke.")

        return retval

    def run(self, user_command, stdout=None, stderr=None, max_retries=0,
            subprocess_timeout=DEFAULT_SUBPROCESS_TIMEOUT):
        """
        Run user_command (blocking) and redirect stdout and/or stderr into
        pre-opened python file descriptor.  When stdout/stderr is not set, the
        output from particular command is ignored.

        :param user_command:
            Command (string) to be executed (note: use shlex.quote).

        :param max_retries:
            When there is ssh connection problem, re-try the action at most
            ``max_retries`` times.  Default is no re-try.  Note that we write
            the output from all the re-tries to the stdout/stderr descriptors
            (when specified).

        :param stdout:
            File descriptor to write standard output into.

        :param stderr:
            File descriptor to write standard error output into.

        :returns:
            Exit status of remote program, or -1 when unexpected failure occurs.

        :type command: str
        :type stdout: file or None
        :type stderr: file or None
        :rtype: list

        """
        rc = -1
        with open(os.devnull, "w", encoding="utf8") as devnull:
            rc = self._retry(self._run, max_retries,
                             user_command, stdout or devnull, stderr or devnull,
                             subprocess_timeout)
        return rc

    def run_expensive(self, user_command, max_retries=0,
                      subprocess_timeout=DEFAULT_SUBPROCESS_TIMEOUT):
        """
        Run user_command (blocking) and return exit status together with
        standard outputs in string variables.  Note that we limit the standard
        (error) output buffer sizes to 10kB; if exceeded, SSHConnectionError is
        raised.  SSHConnectionError is also raised if any non-UTF8 character
        present in one of the outputs.

        :param user_command:
            Command (string) to be run as string (note: use shlex.quote).

        :param max_retries:
            When there is ssh connection problem, re-try the action at most
            ``max_retries`` times.  Default is no re-try.

        :returns:
            Tripple (rc, stdout, stderr)
        """
        return self._retry(self._run_expensive, max_retries, user_command,
                           subprocess_timeout)

    def _run_expensive(self, user_command, subprocess_timeout):
        real_command = self._ssh_base() + [user_command]

        proc = PosixPipedProcess(real_command,
                                 stdout_limit=1024*10,
                                 stderr_limit=1024*10,
                                 timeout=subprocess_timeout)

        stdout = stderr = b''
        for chunk, filename in proc.readchunks():
            if filename == 'stdout':
                stdout += chunk
            else:
                stderr += chunk

        try:
            stdout = stdout.decode("utf-8")
            stderr = stderr.decode("utf-8")
        except UnicodeDecodeError as err:
            raise SSHConnectionError("Non-UTF8 characters in SSH output.") from err

        if proc.timeouted():
            raise SSHConnectionError("SSH timeouted: " +
                                     _user_readable_command(real_command))

        if proc.has_cut():
            raise SSHConnectionError("SSH output was too long: " +
                                     _user_readable_command(real_command))

        if proc.returncode == 255:
            # Value 255 means either that 255 was returned by remote command or
            # the ssh connection broke.  Because we need to handle "Connection
            # broke" issues (resubmit build e.g.), please avoid situations when
            # remote command returns value 255 (or -1).
            raise SSHConnectionError(
                "Connection broke:\nOUT:\n{0}\nERR:\n{1}".format(
                    stdout, stderr))

        return proc.returncode, stdout, stderr

    def _retry(self, method, retries, *args, **kwargs):
        """ Do ``times`` when SSHConnectionError is raised """
        attempt = 0
        while attempt < retries + 1:
            attempt += 1
            try:
                return method(*args, **kwargs)
            except SSHConnectionError as exc:
                sleep = 10
                self.log.error("SSH connection lost on #%s attempt, "
                               "let's retry after %ss, %s", attempt, sleep, exc)
                time.sleep(sleep)
                continue
        raise SSHConnectionError("Unable to finish after {} SSH attempts"
                                 .format(attempt))

    def _full_source_path(self, src):
        """ for easier unittesting """
        host = self.host
        if netaddr.valid_ipv6(host):
            host = "[{}]".format(host)
        return "{}@{}:{}".format(self.user, host, src)

    def rsync_download(self, src, dest, logfile=None, max_retries=0,
                       subprocess_timeout=None, filter_=None):
        """
        Run rsync over pre-allocated socket (by the config)

        :param src:
            Source path on self.host to copy.

        :param dest:
            Destination path on backend to copy ``src` content to.

        :param max_retries:
            When there is ssh connection problem, re-try the action at most
            ``max_retries`` times.  Default is no re-try.

        Store the logs to ``logfile`` within ``dest`` directory.  The dest
        directory needs to exist.
        """
        self._retry(self._rsync_download, max_retries, src, dest, logfile,
                    subprocess_timeout, filter_)

    def _rsync_download(self, src, dest, logfile, subprocess_timeout, filter_):
        ssh_opts = "ssh"
        if self.config_file:
            ssh_opts += " -F " + self.config_file
        if self.port is not None:
            ssh_opts += f" -p {self.port}"

        full_source_path = self._full_source_path(src)

        log_filepath = "/dev/null"
        if logfile:
            log_filepath = os.path.join(dest, logfile)

        command = [
            "/usr/bin/rsync",
            "-rltDvH",
            "--chmod=D755,F644",
            "-e", "'{}'".format(ssh_opts),
        ]
        for value in filter_ or []:
            command.extend(["--filter", shlex.quote(value)])
        command.extend([
            full_source_path,
            "{}/".format(dest),
            "&>", log_filepath,
        ])
        command = " ".join(command)

        self.log.info("rsyncing of %s to %s started", full_source_path, dest)
        with self._popen_timeouted(command, shell=True) as cmd:
            try:
                cmd.wait(timeout=subprocess_timeout)
                self.log.info("rsyncing finished.")
            except Exception as error:
                self.log.error(
                    "Failed to download data from builder due to Popen error, "
                    "original error: %s", error)
                raise SSHConnectionError("Popen failure in rsync.") from error

        if cmd.returncode != 0:
            err_msg = (
                "Failed to download data from builder due to rsync error, "
                "see the rsync log file for details."
            )
            self.log.error(err_msg)
            raise SSHConnectionError(err_msg)



import json
import os
import os.path
import shutil
import time
import traceback

from urllib.request import urlretrieve
from copr.exceptions import CoprRequestException
from munch import Munch

from copr_common.enums import (
    ActionTypeEnum,
    BackendResultEnum,
    StorageEnum,
    CreaterepoReason,
)
from copr_common.worker_manager import WorkerManager
from copr_common.lock import lock

from copr_backend.worker_manager import BackendQueueTask
from copr_backend.storage import storage_for_enum, BackendStorage, PulpStorage

from .sign import create_user_keys, CoprKeygenRequestError
from .constants import PULP_REDIRECT_FILE
from .exceptions import CreateRepoError, CoprSignError, FrontendClientException
from .helpers import (get_redis_logger, silent_remove, ensure_dir_exists,
                      call_copr_repo, copy2_but_hardlink_rpms)
from .sign import get_pubkey


class Action(object):
    """ Object to send data back to fronted

    :param copr_backend.callback.FrontendCallback frontent_callback:
        object to post data back to frontend

    :param destdir: filepath with build results

    :param dict action: dict-like object with action task

    Expected **action** keys:

        - action_type: main field determining what action to apply
        # TODO: describe actions

    """

    @classmethod
    def create_from(cls, opts, action, log=None):
        action_class = cls.get_action_class(action)
        return action_class(opts, action, log)

    @classmethod
    def get_action_class(cls, action):
        action_type = action["action_type"]
        action_class = {
            ActionTypeEnum("legal-flag"): LegalFlag,
            ActionTypeEnum("createrepo"): Createrepo,
            ActionTypeEnum("update_comps"): CompsUpdate,
            ActionTypeEnum("gen_gpg_key"): GenerateGpgKey,
            ActionTypeEnum("rawhide_to_release"): RawhideToRelease,
            ActionTypeEnum("fork"): Fork,
            ActionTypeEnum("remove_dirs"): RemoveDirs,
        }.get(action_type, None)

        if action_type == ActionTypeEnum("delete"):
            object_type = action["object_type"]
            action_class = {
                "copr": DeleteProject,
                "build": DeleteBuild,
                "builds": DeleteMultipleBuilds,
                "chroot": DeleteChroot,
            }.get(object_type, action_class)

        if not action_class:
            raise ValueError("Unexpected action type: {}".format(action))
        return action_class

    # TODO: get more form opts, decrease number of parameters
    def __init__(self, opts, action, log=None):

        self.opts = opts
        self.data = action

        self.ext_data = json.loads(action.get("data", "{}"))

        self.destdir = self.opts.destdir
        self.front_url = self.opts.frontend_base_url
        self.results_root_url = self.opts.results_baseurl

        self.log = log if log else get_redis_logger(self.opts, "backend.actions", "actions")

        self.storage = None
        if isinstance(self.ext_data, dict):
            enum = self.ext_data.get("storage", StorageEnum.backend)
            owner = self.ext_data.get("ownername")
            project = self.ext_data.get("projectname")
            devel = self.ext_data.get("devel")
            appstream = self.ext_data.get("appstream")
            args = [owner, project, appstream, devel, self.opts, self.log]
            self.storage = storage_for_enum(enum, *args)

            # Even though we already have `self.storage` which uses an
            # appropriate storage for the project (e.g. Pulp), the project
            # still has some data on backend (logs, srpm-builds chroot, etc).
            # Many actions need to be performed on `self.storage` and
            # `self.backend_storage` at the same time
            self.backend_storage = storage_for_enum(StorageEnum.backend, *args)

    def __str__(self):
        return "<{}(Action): {}>".format(self.__class__.__name__, self.data)

    def run(self):
        """
        This is an abstract class, implement this function for specific actions
        in their own classes
        """
        raise NotImplementedError()


class LegalFlag(Action):
    def run(self):
        self.log.debug("Action legal-flag: ignoring")


class Createrepo(Action):
    def run(self):
        self.log.info("Action createrepo")
        project_dirnames = self.ext_data["project_dirnames"]
        chroots = self.ext_data["chroots"]
        reason = self.ext_data["reason"]
        result = BackendResultEnum("success")

        for project_dirname in project_dirnames:
            for chroot in chroots:
                success = self.storage.init_project(
                    project_dirname,
                    chroot,
                    reason=reason,
                )
                if not success:
                    result = BackendResultEnum("failure")

        if isinstance(self.storage, PulpStorage):
            self.add_http_redirect()

        return result

    def add_http_redirect(self):
        """
        Create a HTTP redirect for this project.  See:
        https://pagure.io/fedora-infra/ansible/blob/main/f/roles/copr/backend/templates/lighttpd/pulp-redirect.lua.j2
        """
        path = PULP_REDIRECT_FILE
        fullname = "{0}/{1}".format(self.storage.owner, self.storage.project)
        try:
            with open(path, "r", encoding="utf-8") as fp:
                projects = fp.read().splitlines()

            if fullname in projects:
                return

            lockdir = "/var/lock/copr-backend"
            with lock(path, lockdir=lockdir, timeout=-1, log=self.log):
                with open(path, "a", encoding="utf-8") as fp:
                    print(fullname, file=fp)

        except FileNotFoundError:
            # Ignoring because this Copr instance doesn't need redirects
            pass


class GPGMixin(object):
    def generate_gpg_key(self, ownername, projectname):
        if self.opts.do_sign is False:
            # skip key creation, most probably sign component is unused
            return True
        try:
            create_user_keys(ownername, projectname, self.opts)
            return True
        except CoprKeygenRequestError as e:
            self.log.exception(e)
            return False


class Fork(Action, GPGMixin):
    """
    Fork build results from one project to another.

    Please note that CoprDir support is not yet implemented. On a positive note,
    It doesn't polute the main CoprDir with builds from other CoprDirs. The
    results are just not copied.

    See https://github.com/fedora-copr/copr/issues/3820
    """

    def run(self):
        self.log.info("Action fork %s", self.data["object_type"])
        result = BackendResultEnum("success")
        data = json.loads(self.data["data"])
        builds_map = json.loads(self.data["data"])["builds_map"]

        if isinstance(self.storage, PulpStorage):
            # For other actions, we do this in `Action` constructor. But forking
            # actions doesn't have just one owner and project but a source owner
            # and project, and destination owner and project. So we need this
            # workaround to specify which one to work with
            self.storage.owner = self.data["new_value"].split("/")[0]
            self.storage.project = self.data["new_value"].split("/")[1]

        try:
            if self.opts.do_sign:
                pubkey_path = os.path.join(
                    self.destdir,
                    self.data["new_value"],
                    "pubkey.gpg"
                )
                ensure_dir_exists(os.path.dirname(pubkey_path), self.log)

                # Generate brand new gpg key.
                self.generate_gpg_key(data["user"], data["copr"])
                # Put the new public key into forked build directory.
                get_pubkey(data["user"], data["copr"], self.log, self.opts.sign_domain, pubkey_path)

            kwargs = {
                "src_fullname": self.data["old_value"],
                "dst_fullname": self.data["new_value"],
                "builds_map": builds_map,
            }

            # It doesn't matter what storage is used, we need to fork on
            # backend anyway, so that logfiles, etc are copied
            if not self.backend_storage.fork_project(**kwargs):
                result = BackendResultEnum("failure")

            if isinstance(self.storage, PulpStorage):
                if not self.storage.fork_project(**kwargs):
                    result = BackendResultEnum("failure")

        except (CoprSignError, CreateRepoError, CoprRequestException, IOError) as ex:
            self.log.error("Failure during project forking")
            self.log.error(str(ex))
            self.log.error(traceback.format_exc())
            result = BackendResultEnum("failure")
        return result


class DeleteProject(Action):
    def run(self):
        self.log.debug("Action delete copr")
        project_dirnames = self.ext_data["project_dirnames"]

        if not self.storage.owner:
            self.log.error("Received empty ownername!")
            return BackendResultEnum("failure")

        for dirname in project_dirnames:
            if not dirname:
                self.log.warning("Received empty dirname!")
                continue
            self.storage.delete_project(dirname)
            if not isinstance(self.storage, BackendStorage):
                self.backend_storage.delete_project(dirname)

            if isinstance(self.storage, PulpStorage):
                self.remove_http_redirect(dirname)

        return BackendResultEnum("success")

    def remove_http_redirect(self, dirname):
        """
        Remove a HTTP redirect for this project.
        """
        path = PULP_REDIRECT_FILE
        fullname = "{0}/{1}".format(self.storage.owner, dirname)
        try:
            with open(path, "r", encoding="utf-8") as fp:
                projects = fp.read().splitlines()

            if fullname not in projects:
                return

            lockdir = "/var/lock/copr-backend"
            with lock(path, lockdir=lockdir, timeout=-1, log=self.log):
                with open(path, "r", encoding="utf-8") as fp:
                    projects = fp.read().splitlines()

                filtered_projects = [p for p in projects if p != fullname]
                with open(path, "w", encoding="utf-8") as fp:
                    for project in filtered_projects:
                        print(project, file=fp)

        except FileNotFoundError:
            # Ignoring because this Copr instance doesn't need redirects
            pass


class CompsUpdate(Action):
    def run(self):
        self.log.debug("Action comps update")

        ext_data = json.loads(self.data["data"])
        ownername = ext_data["ownername"]
        projectname = ext_data["projectname"]
        chroot = ext_data["chroot"]
        url_path = ext_data["url_path"]

        remote_comps_url = self.opts.frontend_base_url + url_path
        self.log.info(remote_comps_url)

        path = os.path.join(self.destdir, ownername, projectname, chroot)
        ensure_dir_exists(path, self.log)
        local_comps_path = os.path.join(path, "comps.xml")
        result = BackendResultEnum("success")
        if not ext_data.get("comps_present", True):
            silent_remove(local_comps_path)
            self.log.info("deleted comps.xml for %s/%s/%s from %s ",
                          ownername, projectname, chroot, remote_comps_url)
        else:
            try:
                urlretrieve(remote_comps_url, local_comps_path)
                self.log.info("saved comps.xml for %s/%s/%s from %s ",
                              ownername, projectname, chroot, remote_comps_url)
            except Exception:
                self.log.exception("Failed to update comps from %s at location %s",
                                   remote_comps_url, local_comps_path)
                result = BackendResultEnum("failure")
        return result


class DeleteMultipleBuilds(Action):
    def run(self):
        self.log.debug("Action delete multiple builds.")

        # == EXAMPLE DATA ==
        # ownername: @copr
        # projectname: testproject
        # project_dirnames:
        #   testproject:pr:10:
        #     srpm-builds: [00849545, 00849546]
        #     fedora-30-x86_64: [00849545-example, 00849545-foo]
        #   [...]

        project_dirnames = self.ext_data["project_dirnames"]
        build_ids = self.ext_data["build_ids"]

        result = BackendResultEnum("success")
        for project_dirname, chroot_builddirs in project_dirnames.items():
            success = self.storage.delete_builds(
                project_dirname,
                chroot_builddirs,
                build_ids,
            )
            if not success:
                result = BackendResultEnum("failure")
        return result


class DeleteBuild(Action):
    def run(self):
        self.log.info("Action delete build.")

        # == EXAMPLE DATA ==
        # ownername: @copr
        # projectname: TEST1576047114845905086Project10Fork
        # project_dirname: TEST1576047114845905086Project10Fork:pr:12
        # chroot_builddirs:
        #   srpm-builds: [00849545]
        #   fedora-30-x86_64: [00849545-example]

        valid = "object_id" in self.data
        keys = {"ownername", "projectname", "project_dirname",
                "chroot_builddirs", "appstream"}
        for key in keys:
            if key not in self.ext_data:
                valid = False
                break

        if not valid:
            self.log.exception("Invalid action data")
            return BackendResultEnum("failure")

        success = self.storage.delete_builds(
            self.ext_data["project_dirname"],
            self.ext_data["chroot_builddirs"],
            [self.data['object_id']],
        )
        return BackendResultEnum("success" if success else "failure")


class DeleteChroot(Action):
    def run(self):
        self.log.info("Action delete project chroot.")
        chroot = self.ext_data["chrootname"]
        self.storage.delete_repository(chroot)
        return BackendResultEnum("success")


class GenerateGpgKey(Action, GPGMixin):
    def run(self):
        ext_data = json.loads(self.data["data"])
        self.log.info("Action generate gpg key: %s", ext_data)

        ownername = ext_data["ownername"]
        projectname = ext_data["projectname"]

        success = self.generate_gpg_key(ownername, projectname)
        return BackendResultEnum("success") if success else BackendResultEnum("failure")


class RawhideToRelease(Action):
    def run(self):
        data = json.loads(self.data["data"])
        appstream = data["appstream"]
        try:
            chrootdir = os.path.join(self.opts.destdir, data["ownername"],
                                     data["copr_dir"], data["dest_chroot"])
            if not os.path.exists(chrootdir):
                self.log.info("Create directory: %s", chrootdir)
                os.makedirs(chrootdir)

            if isinstance(self.storage, PulpStorage):
                self.storage.init_project(
                    data["copr_dir"],
                    data["dest_chroot"],
                    reason=CreaterepoReason.new_chroot,
                )

            for build in data["builds"]:
                srcdir = os.path.join(self.opts.destdir, data["ownername"],
                                      data["copr_dir"], data["rawhide_chroot"], build)
                destdir = os.path.join(chrootdir, build)
                if os.path.exists(srcdir) and not os.path.exists(destdir):
                    # We can afford doing hardlinks in this case because the
                    # RPMs are not modified at all (contrary to "project
                    # forking", where we have to re-sign the files).
                    self.log.info("Copying directory (link RPMs): %s -> %s",
                                  srcdir, destdir)
                    shutil.copytree(srcdir, destdir,
                                    copy_function=copy2_but_hardlink_rpms)
                    with open(os.path.join(destdir, "build.info"), "a") as f:
                        f.write("\nfrom_chroot={}".format(data["rawhide_chroot"]))

                if isinstance(self.storage, PulpStorage):
                    build_id = int(build.split("-")[0])
                    response =self.storage.client.get_content(
                        [build_id],
                        data["rawhide_chroot"],
                        fields=["prn"],
                    )
                    rpms = response.json()["results"]
                    prns = [rpm["prn"] for rpm in rpms]
                    success = self.storage.create_repository_version(
                        data["copr_dir"],
                        data["dest_chroot"],
                        prns,
                    )
                    if not success:
                        return BackendResultEnum("failure")

            return self._createrepo_repeatedly(chrootdir, appstream)

        # pylint: disable=broad-except
        except Exception:
            self.log.exception("Unexpected error when forking from rawhide")
            return BackendResultEnum("failure")

    def _createrepo_repeatedly(self, chrootdir, appstream):
        # If this project stores RPMs in Pulp, there is no need to run the
        # createrepo on backend, because there are no pacakges there. And we
        # don't need to manually do publish in Pulp either because it does it
        # implicitly.
        if isinstance(self.storage, PulpStorage):
            return BackendResultEnum("success")
        for i in range(5):
            if call_copr_repo(chrootdir, appstream=appstream, logger=self.log):
                return BackendResultEnum("success")
            self.log.error("Createrepo failed, trying again #%s", i)
            time.sleep(10)
        return BackendResultEnum("failure")


class RemoveDirs(Action):
    """
    Delete outdated CoprDir instances.  Frontend gives us only a list of
    sub-directories to remove.
    """
    def _run_internal(self):
        copr_dirs = json.loads(self.data["data"])
        for copr_dir in copr_dirs:
            assert len(copr_dir.split('/')) == 2
            assert ':pr:' in copr_dir
            directory = os.path.join(self.destdir, copr_dir)
            self.log.info("RemoveDirs: removing %s", directory)
            try:
                shutil.rmtree(directory)
            except FileNotFoundError:
                self.log.error("RemoveDirs: %s not found", directory)

    def run(self):
        result = BackendResultEnum("failure")
        try:
            self._run_internal()
            result = BackendResultEnum("success")
        except OSError:
            self.log.exception("RemoveDirs OSError")
        return result


class ActionQueueTask(BackendQueueTask):
    def __init__(self, task):
        self.task = task

    @property
    def id(self):
        return self.task.data["id"]

    @property
    def frontend_priority(self):
        return self.task.data.get("priority", 0)

    @property
    def owner(self):
        """
        Owner of the project this action belongs to
        """
        return self.task.data["project_owner"]


class ActionWorkerManager(WorkerManager):
    worker_prefix = 'action_worker'

    def start_task(self, worker_id, task):
        command = [
            'copr-backend-process-action',
            '--daemon',
            '--task-id', repr(task),
            '--worker-id', worker_id,
        ]
        # TODO: mark as started on FE, and let user know in UI
        self.start_daemon_on_background(command)

    def finish_task(self, worker_id, task_info):
        task_id = self.get_task_id_from_worker_id(worker_id)

        result = Munch()
        result.id = int(task_id)
        result.result = int(task_info['status'])

        try:
            self.frontend_client.update({"actions": [result]})
        except FrontendClientException:
            self.log.exception("can't post to frontend, retrying indefinitely")
            return False
        return True



# coding: utf-8

"""
Test the BuildBackgroundWorker from background_worker_build.py
"""

import copy
import logging
import glob
import json
import os
import shutil
import subprocess
import time
import tempfile
from unittest import mock

from munch import Munch
import pytest

from copr_backend.constants import LOG_REDIS_FIFO
from copr_backend.background_worker_build import (
    BuildBackgroundWorker, MESSAGES, BackendError, _average_step,
)
from copr_backend.job import BuildJob
from copr_backend.exceptions import CoprSignError, FrontendClientException
from copr_backend.vm_alloc import ResallocHost, RemoteHostAllocationTerminated
from copr_backend.background_worker_build import COMMANDS, MIN_BUILDER_VERSION
from copr_backend.sshcmd import SSHConnectionError

import testlib
from testlib import assert_logs_exist, assert_logs_dont_exist
from testlib.repodata import load_primary_xml

# pylint: disable=redefined-outer-name,protected-access

COMMON_MSGS = {
    "not finished": "Switching not-finished job state to 'failed'",
}

def _patch_bw_object(obj, *args, **kwargs):
    return mock.patch("copr_backend.background_worker.{}".format(obj),
                      *args, **kwargs)

def _patch_bwbuild_object(obj, *args, **kwargs):
    return mock.patch("copr_backend.background_worker_build.{}".format(obj),
                      *args, **kwargs)

def _get_rpm_job_object_dict(updated=None):
    job = copy.deepcopy(testlib.VALID_RPM_JOB)
    if updated:
        job.update(updated)
    return job

def _get_rpm_job_object(opts, updated=None):
    job = _get_rpm_job_object_dict(updated)
    return BuildJob(job, opts)

def _get_rpm_job(updated=None):
    job = _get_rpm_job_object_dict(updated)
    response = Munch()
    response.status_code = 200
    response.json = lambda: job
    return response

def _get_srpm_job(updated=None):
    job = copy.deepcopy(testlib.VALID_SRPM_JOB)
    if updated:
        job.update(updated)
    response = Munch()
    response.status_code = 200
    response.json = lambda: job
    return response

def _reset_build_worker():
    worker = BuildBackgroundWorker()
    # Don't waste time with mocking.  We don't want to log anywhere, and we want
    # to let BuildBackgroundWorker adjust the handlers.
    worker.log.handlers = []
    return worker

def _fake_host():
    host = ResallocHost()
    host._is_ready = True
    host.hostname = "1.2.3.4"
    host.ticket = mock.MagicMock(failed=False)
    host.ticket.id = 10
    host.release = mock.MagicMock()
    return host

@pytest.fixture
def f_build_something():
    """
    Prepare worker for RPM or SRPM build
    """
    config = Munch()
    config.workdir = tempfile.mkdtemp(prefix="build-worker-tests-")
    config.be_config_file = testlib.minimal_be_config(config.workdir)

    patchers = [_patch_bw_object("FrontendClient")]
    config.fe_client = patchers[-1].start()

    patchers += [_patch_bwbuild_object("MessageSender")]
    patchers[-1].start()

    config.fe_client_patcher = _patch_bw_object("FrontendClient")
    config.fe_client = config.fe_client_patcher.start()

    config.worker_id = "fake_worker_id_" + str(time.time())

    patchers.append(
        mock.patch("copr_backend.background_worker.sys.argv",
                   ["test-build", "--build-id", "848963",
                    "--worker-id", config.worker_id, "--silent",
                    "--backend-config", config.be_config_file,
                    "--chroot", "fedora-30-x86_64"]),
    )
    patchers[-1].start()

    patchers.append(mock.patch.dict(os.environ, {
        "COPR_BE_CONFIG": config.be_config_file,
        "COPR_TESTSUITE_LOCKPATH": config.workdir,
    }))
    patchers[-1].start()

    config.bw = _reset_build_worker()
    config.bw.redis_set_worker_flag("allocated", "true")

    # Don't waste time with mocking.  We don't want to log anywhere, and we want
    # to let BuildBackgroundWorker adjust the handlers.
    config.bw.log.handlers = []

    patcher = _patch_bwbuild_object("ResallocHostFactory")
    patchers.append(patcher)
    rhf = patcher.start()
    config.host = host = _fake_host()
    rhf.return_value.get_host.return_value = host
    config.resalloc_host_factory = rhf

    config.ssh = testlib.FakeSSHConnection(user="remoteuser", host=host.hostname)
    ssh_patcher = _patch_bwbuild_object("SSHConnection")
    ssh_class = ssh_patcher.start()
    ssh_class.return_value = config.ssh
    patchers.append(ssh_patcher)

    yield config
    for patcher in patchers:
        patcher.stop()
    shutil.rmtree(config.workdir)

    # Clear the adjusted handlers.
    config.bw.log.handlers = []

@pytest.fixture
def f_build_rpm_case_no_repodata(f_build_something):
    """
    Configure the situation when 'copr-backend-process-build' is requested
    to build RPM.
    """
    config = f_build_something
    config.fe_client.return_value.get.return_value = _get_rpm_job()
    yield config

@pytest.fixture
def f_build_srpm(f_build_something):
    """
    Configure the situation when 'copr-backend-process-build' is requested
    to build RPM.
    """
    config = f_build_something
    config.ssh.resultdir = "srpm-builds/02914697"
    config.fe_client.return_value.get.return_value = _get_srpm_job()
    config.ssh.set_command(
        "copr-rpmbuild --verbose --drop-resultdir --srpm --task-url "
        "http://copr-fe/backend/get-srpm-build-task/855954 "
        "--detached",
        0, "666", "",
    )

    patcher = mock.patch(
        "copr_backend.background_worker.sys.argv",
        ["test-build", "--build-id", "848963",
         "--worker-id", config.worker_id, "--silent",
         "--backend-config", config.be_config_file,
         "--chroot", "srpm-builds"],
    )
    patcher.start()
    config.bw = _reset_build_worker()
    yield config
    patcher.stop()

def _create_repodata(in_dir):
    repodata = os.path.join(in_dir, "repodata")
    os.makedirs(repodata)
    repomd = os.path.join(repodata, "repomd.xml")
    with open(repomd, "w"):
        pass


def _create_job_repodata(job):
    _create_repodata(job.chroot_dir)


@pytest.fixture
def f_build_rpm_case(f_build_rpm_case_no_repodata):
    """
    Prepare everything, so the build can succeed.
    """
    config = f_build_rpm_case_no_repodata
    chroot = os.path.join(f_build_rpm_case_no_repodata.workdir, "results",
                          "@copr",
                          "TEST1575431880356948981Project10",
                          "fedora-30-x86_64")
    _create_repodata(chroot)

    config.ssh.set_command(
        "copr-rpmbuild --verbose --drop-resultdir --task-url "
        "http://copr-fe/backend/get-build-task/848963-fedora-30-x86_64 "
        "--chroot fedora-30-x86_64 --detached",
        0, "666", "",
    )
    yield f_build_rpm_case_no_repodata

@pytest.fixture
def f_build_rpm_sign_on(f_build_rpm_case):
    """
    f_build_rpm_case with enabled GPG signing ON
    """
    config = f_build_rpm_case
    with open(config.be_config_file, "a+") as fdconfig:
        fdconfig.write("do_sign=true\n")
        fdconfig.write("keygen_host=keygen.example.com\n")
        fdconfig.write("gently_gpg_sha256=false\n")
    config.bw = _reset_build_worker()
    return config

@_patch_bwbuild_object("time")
def test_waiting_for_repo_fail(mc_time, f_build_rpm_case_no_repodata, caplog):
    """ check that worker loops in _wait_for_repos """
    worker = f_build_rpm_case_no_repodata.bw
    mc_time.time.side_effect = [1, 2, 3, 4, 5, 6, 120, 121]
    worker.process()
    expected = [
        (
            logging.ERROR,
            str(BackendError(MESSAGES["give_up_repo"].format("copr_base"))),
        ),
        (logging.INFO, MESSAGES["repo_waiting"] % "copr_base"),
    ]
    for exp in expected:
        assert exp in [(r[1], r[2]) for r in caplog.record_tuples]

@_patch_bwbuild_object("time")
def test_waiting_for_repo_success(mc_time, f_build_rpm_case_no_repodata, caplog):
    """ check that worker loops in _wait_for_repos """
    worker = f_build_rpm_case_no_repodata.bw

    # on the 6th call to time(), create the repodata
    mc_time.time.side_effect = testlib.TimeSequenceSideEffect(
        [1, 2, 3, 4, 5, 6, 120],
        {6: lambda: _create_job_repodata(worker.job)}
    )

    # shutdown ASAP after _wait_for_repos() call
    def raise_exc():
        raise Exception("duh")
    worker._alloc_host = mock.MagicMock()
    worker._alloc_host.side_effect = raise_exc
    worker.process()

    # _wait_for_repos() succeeded, and we continued to _alloc_host()
    assert len(worker._alloc_host.call_args_list) == 1

    assert (
        logging.INFO, MESSAGES["repo_waiting"] % "copr_base"
    )  in [(r[1], r[2]) for r in caplog.record_tuples]

def test_full_rpm_build_no_sign(f_build_rpm_case, caplog):
    """
    Go through the whole (successful) build of a binary RPM
    """
    worker = f_build_rpm_case.bw
    worker.process()

    results = worker.job.results_dir
    assert os.path.exists(os.path.join(results, "builder-live.log.gz"))
    assert os.path.exists(os.path.join(results, "backend.log.gz"))

    found_success_log_entry = False
    for record in caplog.record_tuples:
        _, level, msg = record
        assert level < logging.ERROR
        if "Finished build: id=848963 failed=False" in msg:
            found_success_log_entry = True
    assert found_success_log_entry

    repodata = load_primary_xml(os.path.join(results, "..", "repodata"))
    assert repodata["names"] == {"example"}
    assert repodata["packages"]["example"]["href"] == \
        "00848963-example/example-1.0.14-1.fc30.x86_64.rpm"

    assert worker.job.built_packages == "example 1.0.14"
    assert_messages_sent(["build.start", "chroot.start", "build.end"], worker.sender)

def test_prev_build_backup(f_build_rpm_case):
    worker = f_build_rpm_case.bw
    worker.process()
    worker.process()
    prev_results = os.path.join(worker.job.results_dir, "prev_build_backup")
    assert glob.glob(os.path.join(prev_results, '*.rpm')) == []
    assert os.path.exists(os.path.join(prev_results, "builder-live.log.gz"))
    assert os.path.exists(os.path.join(prev_results, "backend.log.gz"))
    rsync_patt = "*{}.rsync.log".format(worker.job.build_id)
    assert len(glob.glob(os.path.join(prev_results, rsync_patt))) == 1


def test_full_srpm_build(f_build_srpm):
    worker = f_build_srpm.bw
    worker.process()
    assert worker.job.pkg_name == "hello"

    # TODO: fix this is ugly pkg_version testament
    assert worker.job.pkg_version is None
    assert worker.job.__dict__["pkg_version"] == "2.8-1"

    assert worker.job.srpm_url == (
        "https://example.com/results/@copr/PROJECT_2/srpm-builds/"
        "00855954/hello-2.8-1.src.rpm")


@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
@mock.patch("copr_backend.sign._sign_one")
def test_build_and_sign(mc_sign_one, f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw
    worker.process()
    pub_key = os.path.join(worker.job.destdir, "pubkey.gpg")
    with open(pub_key, "r") as pub:
        content = pub.readlines()
        assert content == ["fake pub key content\n"]

    mail = "@copr#TEST1575431880356948981Project10@copr.fedorahosted.org"
    rpm = os.path.join(worker.job.results_dir,
                       "example-1.0.14-1.fc30.x86_64.rpm")
    srpm = os.path.join(worker.job.results_dir,
                        "example-1.0.14-1.fc30.src.rpm")
    expected_calls = [
        mock.call(rpm, mail, "sha256", mc_sign_one.call_args_list[0][0][3]),
        mock.call(srpm, mail, "sha256", mc_sign_one.call_args_list[1][0][3]),
    ]
    for call in expected_calls:
        assert call in mc_sign_one.call_args_list
    assert len(mc_sign_one.call_args_list) == 2
    for record in caplog.record_tuples:
        _, level, _ = record
        assert level <= logging.INFO

@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
@mock.patch("copr_backend.sign._sign_one")
@_patch_bwbuild_object("sign_rpms_in_dir")
def test_sign_built_packages_exception(mc_sign_rpms, mc_sign_one,
                                       f_build_rpm_sign_on, caplog):
    _side_effect = mc_sign_one
    mc_sign_rpms.side_effect = CoprSignError("test")
    config = f_build_rpm_sign_on
    worker = config.bw
    worker.process()
    messages = [
        (logging.ERROR, "Copr GPG signing problems: test"),
    ]
    found_fail = False
    for msg in messages:
        assert msg in [(r[1], r[2]) for r in caplog.record_tuples]
    for msg in caplog.record_tuples:
        _, _, text = msg
        if "Finished build: id=848963 failed=True" in text:
            found_fail = True
    assert found_fail

def _get_log_content(job, log="backend.log.gz"):
    logfile = os.path.join(job.results_dir, log)
    cmd = ["gunzip", "-c", logfile]
    return subprocess.check_output(cmd).decode("utf-8")

def test_unexpected_exception(f_build_rpm_case, caplog):
    config = f_build_rpm_case
    worker = config.bw
    def _raise():
        raise IOError("blah")

    worker._check_vm = _raise
    worker.process()

    redis = worker._redis
    log_entry = None

    # check that the traceback is logged to worker.log
    while True:
        log_entry = redis.rpop(LOG_REDIS_FIFO)
        if not log_entry:
            break
        if "Traceback" in log_entry:
            break

    log_dict = json.loads(log_entry)
    msg = "Unexpected exception\nTraceback (most recent call last)"
    assert msg in log_dict["msg"]

    content = _get_log_content(worker.job)
    assert "Traceback" not in content

    found_line = None
    for line in content.splitlines():
        if "Unexpected exception" in line:
            found_line = line
            break

    # check that the shortened variant is in log
    assert "/test_background_worker_build.py:" in found_line

def test_build_info_file_failure(f_build_rpm_case):
    config = f_build_rpm_case
    worker = config.bw
    worker.job = _get_rpm_job_object(worker.opts)
    os.makedirs(worker.job.results_dir)
    info_file = os.path.join(worker.job.results_dir, "build.info")
    worker.host = Munch()
    worker.host.hostname = "0.0.0.0"
    worker._fill_build_info_file()
    with open(info_file, "r") as info_fd:
        content = info_fd.readlines()
        # TODO: fix missing newline
        assert content == ['build_id=848963\n', 'builder_ip=0.0.0.0']
    # make it "non-writable"
    os.unlink(info_file)
    os.mkdir(info_file)
    with pytest.raises(BackendError) as err:
        worker._fill_build_info_file()
    assert "Backend process error: Can't write to " in str(err.value)
    assert "00848963-example/build.info'" in str(err.value)

def test_invalid_job_info(f_build_rpm_case, caplog):
    config = f_build_rpm_case
    worker = config.bw
    get = config.fe_client.return_value.get.return_value
    job = get.json()
    del job["chroot"]
    worker.process()
    assert_logs_exist([
        "Backend process error: Frontend job doesn't provide chroot",
        COMMON_MSGS["not finished"],
    ], caplog)
    assert_logs_dont_exist([
        "took None",
    ], caplog)

@mock.patch("copr_backend.vm_alloc.time.sleep", mock.MagicMock())
@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_cancel_build_on_vm_allocation(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw

    # let it think it is started by "worker manager"
    class _CheckReady:
        attempt = 0
        def __call__(self):
            # process() configures logging, so drop the loggers to avoid
            # an ugly test output
            self.attempt += 1
            if self.attempt > 3:
                # Deliver cancel request to checker thread (this is
                # delivered by WorkerManager normally).
                worker.redis_set_worker_flag("cancel_request", 1)

            # When the "cancel_request" is processed, we get the "canceling"
            # response.  Resalloc client would normally recognize that the
            # ticket is closed, and raised RemoteHostAllocationTerminated.
            if worker.redis_get_worker_flag("canceling"):
                raise RemoteHostAllocationTerminated
            return False

    config.host._is_ready = False
    config.host.check_ready = mock.MagicMock()
    config.host.check_ready.side_effect = _CheckReady()

    found_records = {}
    def _find_records(record_msg):
        msgs = [
            "Build was canceled", # cancel handled
            COMMON_MSGS["not finished"],
            "Worker failed build", # needs to finish
            "Unable to compress file", # there's no builder-live.log yet
        ]
        for msg in msgs:
            if not msg in found_records:
                found_records[msg] = False
            if msg in record_msg:
                found_records[msg] = True

    worker.process()
    for record in caplog.record_tuples:
        _, _, msg = record
        _find_records(msg)
    for key, value in found_records.items():
        assert (key, value) == (key, True)
    assert worker.job.status == 0  # failure

class _CancelFunction():
    def __init__(self, worker):
        self.worker = worker
    def __call__(self):
        # request cancelation
        self.worker.redis_set_worker_flag("cancel_request", 1)
        # and do something till _cancel_vm_allocation() doesn't let us know
        while self.worker.redis_get_worker_flag("canceling") is None:
            time.sleep(0.25)

@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_cancel_build_on_tail_log_no_ssh(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw

    config.ssh.set_command(
        "copr-rpmbuild-log",
        0, "canceled stdout\n", "canceled stderr\n",
        _CancelFunction(worker),
    )
    worker.process()
    exp_msgs = {
        "Can't ssh to cancel build.",
        "Build was canceled",
        COMMON_MSGS["not finished"],
        "Worker failed build",
    }
    found_messages = set()

    for record in caplog.record_tuples:
        _, _, msg = record
        for exp_msg in exp_msgs:
            if exp_msg in msg:
                found_messages.add(exp_msg)
    assert exp_msgs == found_messages
    log = _get_log_content(worker.job, "builder-live.log.gz")
    assert "canceled stdout" in log

@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_cancel_before_vm(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw
    # set this early
    worker.redis_set_worker_flag("cancel_request", 1)
    worker.process()
    assert_logs_exist([
        "Build was canceled",
        "Canceling the build early",
        COMMON_MSGS["not finished"],
        "Worker failed build",
    ], caplog)
    assert_logs_dont_exist(["Releasing VM back to pool"], caplog)

@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_cancel_before_start(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw

    # cancel request right before starting the build
    worker._fill_build_info_file = mock.MagicMock()
    worker._fill_build_info_file.side_effect = \
        lambda: worker.redis_set_worker_flag("cancel_request", 1)

    worker.process()
    assert_logs_exist([
        "Build was canceled",
        "Releasing VM back to pool",
        "Canceling the build early",
        COMMON_MSGS["not finished"],
        "Worker failed build",
    ], caplog)

@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_build_retry(f_build_rpm_sign_on):
    config = f_build_rpm_sign_on
    worker = config.bw
    class _SideEffect():
        counter = 0
        def __call__(self):
            self.counter += 1
            if self.counter < 2:
                return (1, "out", "err")
            if self.counter < 3:
                return (0, "0.38", "")
            return (0, MIN_BUILDER_VERSION, "")

    config.ssh.set_command(
        COMMANDS["rpm_q_builder"],
        1, "err stdout\n", "err stderr\n",
        return_action=_SideEffect())

    worker.process()
    log = _get_log_content(worker.job)
    find_msgs = {
        MESSAGES["copr_rpmbuild_missing"].format("err"),
        "Minimum version for builder is " + MIN_BUILDER_VERSION,
        "Allocating ssh connection to builder",
        "Finished build: id=848963 failed=False",
    }
    found_msgs = set()
    for line in log.splitlines():
        for find in find_msgs:
            if find in line:
                found_msgs.add(find)
    assert find_msgs == found_msgs
    assert _get_log_content(worker.job, "builder-live.log.gz").splitlines() == \
        ["build log stdout", "build log stderr"]

def assert_messages_sent(topics, sender):
    """ check msg bus calls """
    assert len(topics) == len(sender.announce.call_args_list)
    for topic in topics:
        found = False
        for call in sender.announce.call_args_list:
            if topic == call[0][0]:
                found = True
        assert (found, topic) == (True, topic)

def test_fe_disallowed_start(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw
    config.fe_client.return_value.starting_build.return_value = False
    worker.process()
    assert any(["Frontend forbade to start" in r[2] for r in caplog.record_tuples])
    assert any(["Worker failed build" in r[2] for r in caplog.record_tuples])

def test_fe_failed_start(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw
    job = _get_rpm_job()
    job.status_code = 403
    config.fe_client.return_value.get.side_effect =\
        FrontendClientException("Error message from SafeRequest")
    worker.process()
    assert_logs_exist([
        "Failed to download build info: Error message from SafeRequest",
        "Backend process error: Failed to get the build task"
        " get-build-task/848963-fedora-30-x86_64",
        "No job object from Frontend",
    ], caplog)
    # check that worker manager is notified
    assert worker.redis_get_worker_flag("status") == "done"

@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_cancel_script_failure(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw
    config.ssh.set_command(
        "copr-rpmbuild-log",
        0, "canceled stdout\n", "canceled stderr\n",
        _CancelFunction(worker),
    )
    config.ssh.set_command(
        "copr-rpmbuild-cancel",
        1, "output", "err output",
    )
    worker.process()
    assert_logs_exist([
        "Can't cancel build\nout:\noutput\nerr:\nerr output",
        "Build was canceled",
        COMMON_MSGS["not finished"],
        "Worker failed build, took",
    ], caplog)

@_patch_bwbuild_object("CANCEL_CHECK_PERIOD", 0.5)
@mock.patch("copr_backend.sign.SIGN_BINARY", "tests/fake-bin-sign")
def test_cancel_build_during_log_download(f_build_rpm_sign_on, caplog):
    config = f_build_rpm_sign_on
    worker = config.bw
    config.ssh.set_command(
        "copr-rpmbuild-log",
        0, "canceled stdout\n", "canceled stderr\n",
        _CancelFunction(worker),
    )
    config.ssh.set_command("copr-rpmbuild-cancel", 0, "out", "err")
    worker.process()
    assert_logs_exist([
        "Cancel request succeeded\nout:\nouterr:\nerr",
        "Build was canceled",
        COMMON_MSGS["not finished"],
    ], caplog)

def test_ssh_connection_error(f_build_rpm_case, caplog):
    class _SideEffect:
        counter = 0
        def __call__(self):
            self.counter += 1
            if self.counter == 1:
                return (1, b"err stdout", "err stderr")
            return (0, "", "")

    config = f_build_rpm_case
    ssh = config.ssh
    ssh.set_command("copr-builder-ready fedora-30-x86_64",
                    0, "", "", return_action=_SideEffect())
    worker = config.bw
    worker.process()
    assert_logs_exist([
        "Retry #1 (on other host)",
        "Worker succeeded build",
    ], caplog)

def test_average_step():
    assert _average_step([]) == float("inf")
    assert _average_step([1, 2, 3]) == float("inf")
    assert _average_step([1, 2, 3, 4, 6]) == 1.25

@_patch_bwbuild_object("time.sleep", mock.MagicMock())
@_patch_bwbuild_object("time.time")
@_patch_bwbuild_object("BuildBackgroundWorker._build_timeouted")
def test_retry_for_ssh_tail_failure(mc_timeouted, mc_time, f_build_rpm_case,
                                    caplog):
    mc_timeouted.return_value = False
    mc_time.side_effect = list(range(500))
    class _SideEffect:
        counter = 0
        def __call__(self):
            self.counter += 1
            if self.counter > 5:
                return (0, "", "")
            raise SSHConnectionError("test failure")
    config = f_build_rpm_case
    ssh = config.ssh
    ssh.set_command("copr-rpmbuild-log",
                    0, "", "", return_action=_SideEffect())
    worker = config.bw
    worker.process()
    assert_logs_exist([
        "Retry #1 (on other host)",
        "Worker succeeded build",
        "Giving up for unstable SSH",
    ], caplog)
    assert_messages_sent(["build.start", "chroot.start", "build.end"], worker.sender)

def test_build_failure(f_build_rpm_case, caplog):
    config = f_build_rpm_case
    config.ssh.unlink_success = True
    worker = config.bw
    worker.process()
    assert_logs_exist([
        "Backend process error: No success file => build failure",
        "Worker failed build, took ",
        "Finished build: id=848963 failed=True ",
    ], caplog)
    assert_messages_sent(["build.start", "chroot.start", "build.end"], worker.sender)

@mock.patch("copr_backend.storage.call_copr_repo")
def test_createrepo_failure(mc_call_copr_repo, f_build_rpm_case, caplog):
    mc_call_copr_repo.return_value = False
    config = f_build_rpm_case
    worker = config.bw
    worker.process()
    assert_logs_exist([
        "Backend process error: createrepo failed",
        "Worker failed build, took ",
        "Finished build: id=848963 failed=True ",
    ], caplog)

def test_existing_compressed_file(f_build_rpm_case, caplog):
    config = f_build_rpm_case
    config.ssh.precreate_compressed_log_file = True
    worker = config.bw
    worker.process()
    assert_logs_exist([
        "Worker succeeded build, took ",
        "builder-live.log.gz exists",
        "Finished build: id=848963 failed=False ",  # still success!
    ], caplog)

def test_tail_f_nonzero_exit(f_build_rpm_case, caplog):
    config = f_build_rpm_case
    worker = config.bw
    class _SideEffect:
        counter = 0
        def __call__(self):
            self.counter += 1
            if self.counter > 3:
                return (0, "ok\n", "ok\n")
            return (1, "fail out\n", "fail err\n")
    config.ssh.set_command(
        "copr-rpmbuild-log",
        0, "failed stdout\n", "failed stderr\n",
        return_action=_SideEffect(),
    )
    worker.process()
    assert_logs_exist([
        "Retry #3 (on other host)",
        "Worker succeeded build, took ",
        "Finished build: id=848963 failed=False ",  # still success!
    ], caplog)

def test_ignored_copr_rpmbuild_daemon_output(f_build_srpm, caplog):
    config = f_build_srpm
    config.ssh.set_command(
        "copr-rpmbuild --verbose --drop-resultdir --srpm --task-url "
        "http://copr-fe/backend/get-srpm-build-task/855954 "
        "--detached",
        0, "unused stdout", "unused stderr",
    )
    config.bw.process()
    assert_logs_exist([
        "The copr-rpmbuild seems started, per:\n"
        "stdout: unused stdout\n"
        "stderr: unused stderr\n"
    ], caplog)
    assert_logs_dont_exist([
        "Retry",
    ], caplog)

def test_unable_to_start_builder(f_build_srpm, caplog):
    config = f_build_srpm
    config.ssh.set_command(
        "copr-rpmbuild --verbose --drop-resultdir --srpm --task-url "
        "http://copr-fe/backend/get-srpm-build-task/855954 "
        "--detached",
        10, "stdout\n", "stderr\n",
    )
    config.bw.process()
    assert_logs_exist([
        "Can't start copr-rpmbuild",
        "out:\nstdout\nerr:\nstderr\n",
        "builder-live.log: No such file or directory",
    ], caplog)
    assert_logs_dont_exist(["Retry"], caplog)

@_patch_bwbuild_object("time.sleep", mock.MagicMock())
def test_retry_vm_factory_take(f_build_srpm, caplog):
    config = f_build_srpm
    rhf = config.resalloc_host_factory
    host = config.host
    fake_host = Munch()
    fake_host.wait_ready = lambda: False
    fake_host.release = lambda: None
    fake_host.info = "fake host"
    rhf.return_value.get_host.side_effect = [fake_host, host]
    config.bw.process()
    assert_logs_exist([
        "VM allocation failed, trying to allocate new VM",
        "Finished build: id=855954 failed=False",
    ], caplog)
    assert config.bw.job.status == 1  # success

def test_failed_build_retry(f_build_rpm_case, caplog):
    config = f_build_rpm_case
    rhf = config.resalloc_host_factory
    hosts = [_fake_host() for _ in range(4)]
    for index in range(4):
        hosts[index].hostname = "1.2.3." + str(index)
    rhf.return_value.get_host.side_effect = hosts
    ssh = config.ssh
    ssh.set_command("copr-builder-ready fedora-30-x86_64",
                    1, "", "not found")

    config.bw.process()
    assert_logs_exist([
        "Three host tried without success: {'1.2.3.",
        COMMON_MSGS["not finished"],
        "Worker failed build, took ",
    ], caplog)
    assert config.bw.job.status == 0
    # Only build.end sent by this worker (after reset)
    assert_messages_sent(["build.end"], config.bw.sender)

def test_buildjob_tags(f_build_rpm_case):
    config = f_build_rpm_case
    worker = config.bw
    worker.job = _get_rpm_job_object(worker.opts)
    assert worker.job.tags == ['arch_x86_64', 'test_tag']

def test_task_url(f_build_rpm_case):
    config = f_build_rpm_case
    worker = config.bw
    worker.job = _get_rpm_job_object(worker.opts)
    assert worker.job.task_url == \
        "http://copr-fe/backend/get-build-task/848963-fedora-30-x86_64"

def test_buildjob_chroot_dir(f_build_rpm_case):
    config = f_build_rpm_case
    worker = config.bw
    job_dict = copy.deepcopy(testlib.VALID_SUBPROJECT_PRM_JOB)
    job = BuildJob(job_dict, worker.opts)
    assert job.chroot_dir.endswith("copr-pull-requests:pr:3568/fedora-40-x86_64")



"""
Support for various data storages, e.g. results directory on backend, Pulp, etc.
"""

import os
import gc
from tempfile import TemporaryDirectory
from concurrent.futures import ThreadPoolExecutor, as_completed
# We need to stop using the deprecated distutils module.
# See https://peps.python.org/pep-0632/
# As of Python 3.12, it doesn't even exist now, we just use an alias provided
# by python3-setutpools, see
# https://fedoraproject.org/wiki/Changes/Python3.12#The_Python_standard_library_distutils_module_will_be_removed
from distutils.dir_util import copy_tree  # pylint: disable=deprecated-module
from distutils.errors import DistutilsFileError # pylint: disable=deprecated-module
import shutil
import requests
from copr_common.enums import StorageEnum, CreaterepoReason
from copr_common.request import RequestError, SafeRequest
from copr_backend.helpers import call_copr_repo, build_chroot_log_name, ensure_dir_exists
from copr_backend.pulp import PulpClient
from copr_backend.exceptions import CoprBackendError
from .sign import resign_rpms_in_dir


def storage_for_job(job, opts, log):
    """
    Return an appropriate storage object for a given job
    """
    # pylint: disable=too-many-function-args
    return storage_for_enum(
        job.storage,
        job.project_owner,
        job.project_name,
        job.appstream,
        job.uses_devel_repo,
        opts,
        log,
    )


def storage_for_enum(enum_value, owner, project, appstream, devel, opts, log):
    """
    Return an appropriate `StorageEnum` value
    """
    args = [owner, project, appstream, devel, opts, log]
    if enum_value == StorageEnum.pulp:
        return PulpStorage(*args)
    return BackendStorage(*args)


class Storage:
    """
    Storage agnostic, high-level interface for storing and acessing our data
    """

    def __init__(self, owner, project, appstream, devel, opts, log):
        self.owner = owner
        self.project = project
        self.appstream = appstream
        self.devel = devel
        self.opts = opts
        self.log = log

    def init_project(self, dirname, chroot, reason=None):
        """
        Make sure users can enable a DNF repository for this project/chroot

        This function is called in various situations.
        Some of them you might not expect:

        1. A new project is created
        2. User clicks the "Regenerate" button in project overview
        3. The "manual createrepo" feature is toggled for the project
        4. A chroot is enabled for the project
        """
        raise NotImplementedError

    def upload_build_results(self, rpm_paths, chroot, max_workers=1, build_id=None):
        """
        Add results for a new build to the storage
        """

    def publish_repository(self, chroot, **kwargs):
        """
        Publish new build results in the repository

        See when we run createrepo
        https://docs.pagure.org/copr.copr/createrepo.html
        """
        raise NotImplementedError

    def delete_repository(self, chroot):
        """
        Delete a repository and all of its builds
        """
        raise NotImplementedError

    def delete_project(self, dirname):
        """
        Delete the whole project and all of its repositories and builds
        """
        raise NotImplementedError

    def delete_builds(self, dirname, chroot_builddirs, build_ids):
        """
        Delete multiple builds from the storage
        """
        raise NotImplementedError

    def repository_exists(self, dirname, chroot, baseurl):
        """
        Does a repository exist?
        """
        raise NotImplementedError

    def fork_project(self, src_fullname, dst_fullname, builds_map):
        """
        Fork build results from one project to another
        """

    def find_build_results(self, results_dir):
        """
        Return a list of absolute paths of all RPMs within this directory.
        The results are returned in an undetermined order.
        """
        results = []
        for root, _, files in os.walk(results_dir):
            for name in files:
                if os.path.basename(root) == "prev_build_backup":
                    continue

                if not name.endswith(".rpm"):
                    continue

                path = os.path.join(root, name)
                results.append(path)
        return results


class BackendStorage(Storage):
    """
    Store build results in `/var/lib/copr/public_html/results/`
    """

    def init_project(self, dirname, chroot, reason=None):
        self.log.info("Creating repo for: %s/%s/%s",
                      self.owner, dirname, chroot)
        repo = os.path.join(self.opts.destdir, self.owner,
                            dirname, chroot)
        try:
            os.makedirs(repo)
            self.log.info("Empty repo so far, directory created")
        except FileExistsError:
            pass

        return call_copr_repo(repo, appstream=self.appstream, devel=self.devel,
                              logger=self.log)

    def publish_repository(self, chroot, **kwargs):
        assert "chroot_dir" in kwargs
        assert "target_dir_name" in kwargs

        base_url = "/".join([self.opts.results_baseurl, self.owner,
                             self.project, chroot])

        self.log.info("Incremental createrepo run, adding %s into %s, "
                      "(auto-create-repo=%s)", kwargs["target_dir_name"],
                      base_url, not self.devel)
        return call_copr_repo(kwargs["chroot_dir"], devel=self.devel,
                              add=[kwargs["target_dir_name"]],
                              logger=self.log,
                              appstream=self.appstream)

    def delete_repository(self, chroot):
        chroot_path = os.path.join(
            self.opts.destdir, self.owner, self.project, chroot)

        self.log.info("Going to delete: %s", chroot_path)
        if not os.path.isdir(chroot_path):
            self.log.error("Directory %s not found", chroot_path)
            return
        shutil.rmtree(chroot_path)

    def delete_project(self, dirname):
        path = os.path.join(self.opts.destdir, self.owner, dirname)
        if os.path.exists(path):
            self.log.info("Removing copr dir %s", path)
            shutil.rmtree(path)

    def delete_builds(self, dirname, chroot_builddirs, build_ids):
        result = True
        for chroot, subdirs in chroot_builddirs.items():
            chroot_path = os.path.join(
                self.opts.destdir, self.owner, dirname, chroot)
            if not os.path.exists(chroot_path):
                self.log.error("%s chroot path doesn't exist", chroot_path)
                result = False
                continue

            self.log.info("Deleting subdirs [%s] in %s",
                          ", ".join(subdirs), chroot_path)

            # Run createrepo first and then remove the files (to avoid old
            # repodata temporarily pointing at non-existing files)!
            # In srpm-builds we don't create repodata at all
            if chroot != "srpm-builds":
                repo = call_copr_repo(
                    chroot_path, delete=subdirs, devel=self.devel,
                    appstream=self.appstream, logger=self.log)
                if not repo:
                    result = False

            for build_id in build_ids or []:
                log_paths = [
                    os.path.join(chroot_path, build_chroot_log_name(build_id)),
                    # we used to create those before
                    os.path.join(chroot_path, 'build-{}.rsync.log'.format(build_id)),
                    os.path.join(chroot_path, 'build-{}.log'.format(build_id))]
                for log_path in log_paths:
                    try:
                        os.unlink(log_path)
                    except OSError:
                        self.log.debug("can't remove %s", log_path)
        return result

    def repository_exists(self, dirname, chroot, baseurl):
        repodata = os.path.join(self.opts.destdir, self.owner, dirname,
                                chroot, "repodata", "repomd.xml")
        self.log.info("Checking that %s exists", repodata)
        return os.path.exists(repodata)

    def fork_project(self, src_fullname, dst_fullname, builds_map):
        dst_owner, dst_project = dst_fullname.split("/")
        old_path = os.path.join(self.opts.destdir, src_fullname)
        new_path = os.path.join(self.opts.destdir, dst_fullname)

        if not os.path.exists(old_path):
            self.log.info("Source copr directory doesn't exist: %s", old_path)
            return False

        chroot_paths = set()
        for chroot, src_dst_dir in builds_map.items():
            if not chroot or not src_dst_dir:
                continue

            for old_dir_name, new_dir_name in src_dst_dir.items():
                src_dir, dst_dir = old_dir_name, new_dir_name

                if not src_dir or not dst_dir:
                    continue

                new_chroot_path = self._fork_build(
                    chroot,
                    old_path,
                    new_path,
                    src_dir,
                    dst_dir,
                    dst_owner,
                    dst_project,
                )
                if new_chroot_path and chroot != "srpm-builds":
                    chroot_paths.add(new_chroot_path)

        for chroot_path in chroot_paths:
            if not call_copr_repo(chroot_path, logger=self.log):
                return False
        return True

    def _fork_build(self, chroot, old_path, new_path, src_dir, dst_dir, dst_owner, dst_project):
        # pylint: disable=too-many-positional-arguments
        old_chroot_path = os.path.join(old_path, chroot)
        new_chroot_path = os.path.join(new_path, chroot)

        src_path = os.path.join(old_chroot_path, src_dir)
        dst_path = os.path.join(new_chroot_path, dst_dir)

        ensure_dir_exists(dst_path, self.log)

        try:
            copy_tree(src_path, dst_path)
        except DistutilsFileError as e:
            self.log.error(str(e))
            return None

        resign_rpms_in_dir(dst_owner, dst_project, dst_path, chroot, self.opts, self.log)

        self.log.info("Forked build %s as %s", src_path, dst_path)
        return new_chroot_path


class PulpStorage(Storage):
    """
    Store build results in Pulp
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = PulpClient.create_from_config_file(log=self.log, opts=self.opts)

    def init_project(self, dirname, chroot, reason=None):
        repository_name = self._repository_name(chroot, dirname)
        try:
            response = self.client.create_repository(repository_name)
        except RequestError as ex:
            if "This field must be unique" not in ex.response.text:
                self.log.error(
                    "Failed to create a Pulp repository %s because of %s",
                    repository_name,
                    ex.response.text,
                )
                return False

        # When a repository is mentioned in other endpoints, it needs to be
        # mentioned by its href, not name
        response = self.client.get_repository(repository_name)
        repository = response.json()["results"][0]
        repository_href = repository["pulp_href"]

        distribution_name = self._distribution_name(chroot, dirname)
        public_distribution_name = self._distribution_name(chroot, dirname, devel=False)
        devel_distribution_name = self._distribution_name(chroot, dirname, devel=True)

        try:
            response = self.client.create_distribution(
                distribution_name,
                repository_href,
            )
        except RequestError as ex:
            if "This field must be unique" not in ex.response.text:
                self.log.error(
                    "Failed to create a Pulp distribution %s because of %s",
                    public_distribution_name,
                    ex.response.text,
                )
                return False

        # If a project enabled the manual createrepo mode, we need to create a
        # devel distribution to be consumed by other builds within the project
        if self.devel:
            # Wait until we can get the publication
            if task := response.json().get("task"):
                self.client.wait_for_finished_task(task)
            response = self.client.get_publication(repository_href)
            publication = response.json()["results"][0]["pulp_href"]
            public_distribution = self._get_distribution(chroot, dirname, devel=False)

            response = self.client.update_distribution(public_distribution, publication=publication)
            if not response.ok:
                self.log.error("Failed to update Pulp distribution %s for because %s",
                               public_distribution, response.text)

        # We want to disable the manual createrepo feature on a project
        elif reason == CreaterepoReason.manual_createrepo_toggle:
            public_distribution = self._get_distribution(chroot)
            response = self.client.update_distribution(
                public_distribution,
                repository=repository_href,
            )
            if not response.ok:
                self.log.error("Failed to update Pulp distribution %s for because %s",
                               public_distribution_name, response.text)
                return False

        # The "Regenerate" button was clicked for a manual createrepo project.
        # We want to copy packages from the devel repo to the public repo.
        # Or more precisely, point the public distribution to
        # the same publication as the devel distribution points to.
        #
        # A note regarding implementation:
        # It is weird to do this in the `init_project` method, pointing to the
        # fact that our abstractions are probably wrong and should be fixed.
        # We have the `publish_repository` method which is called after a build
        # is finished. We don't want to copy the packages there. This
        # `init_project` method is called when a project is created (for which
        # the copying does nothing) but also when a "Regenerate" button is
        # clicked (for which we copy the data). We should probably separate
        # these two concepts.
        elif reason == CreaterepoReason.manual_createrepo_event:
            response = self.client.get_publication(repository_href)
            publication = response.json()["results"][0]["pulp_href"]
            public_distribution = self._get_distribution(chroot, full=True)

            # Only update the Pulp distribution if the project uses the manual
            # createrepo feature. If not, do nothing.
            # We cannot rely on the `self.devel` option because doesn't say
            # anything about the project setting of the feature, it only says
            # that the current action should affect the public repository. If
            # the distribution points to a publication, than the project uses
            # manual createrepo, if it points to a repository, it is a normal
            # project.
            if public_distribution["publication"]:
                self.log.info(
                    "Pointing %s to the same publication as %s which is %s",
                    public_distribution_name,
                    devel_distribution_name,
                    publication,
                )
                response = self.client.update_distribution(
                    public_distribution["pulp_href"],
                    publication=publication,
                )
                if not response.ok:
                    self.log.error("Failed to update Pulp distribution %s for because %s",
                                   public_distribution_name, response.text)
                    return False

        # And finally, run the actual createrepo for either the devel or
        # the public repository
        response = self.client.create_publication(repository_href)
        return response.ok

    def upload_rpm(self, path, labels):
        """
        Add an RPM to the storage
        """
        i = 1
        while True:
            try:
                response = self.client.create_content(path, labels)
                if response.ok:
                    return response
                self.log.error(
                    "Unsuccessful response when creating Pulp content for "
                    "%s, %s %s (attempt #%s)",
                    path, response.reason, response.text, i,
                )
            except RequestError as ex:
                self.log.error(
                    "Failed to create Pulp content for: %s, %s %s (attempt #%s)",
                    path, ex.response.reason, ex.response.text, i,
                )
                # We need to explicitly the force garbage collector to run
                # right now, otherwise we will leak a lot of memory by trying
                # to upload the (large) file again and again. Which means
                # reading it again and again. Not sure why the previous memory
                # couldn't be re-used and not sure why the garbage collector
                # doesn't free it automatically. It doesn't though, this fixes
                # a real memory leak that led to the process being OOM-killed.
                gc.collect()
            i += 1

    def upload_build_results(self, rpm_paths, chroot, max_workers=1, build_id=None):
        futures = {}
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for path in rpm_paths:
                labels = {"build_id": build_id, "chroot": chroot}
                futures[executor.submit(self.upload_rpm, path, labels)] = path

            failed_uploads = []
            exceptions = []
            uploaded = {}
            for i, future in enumerate(as_completed(futures), start=1):
                filepath = futures[future]
                try:
                    response = future.result()
                    if response.ok:
                        created = response.json().get("pulp_href")
                        uploaded[os.path.basename(filepath)] = {
                            "build_id": build_id,
                            "path": filepath,
                            "pulp_href": created,
                        }
                        self.log.info(
                            "[%s/%s] Uploaded to Pulp: %s",
                            i, len(futures), filepath,
                        )
                    else:
                        failed_uploads.append(filepath)
                except RuntimeError as exc:
                    exceptions.append(f"{filepath} generated an exception: {exc}")

            if failed_uploads:
                raise CoprBackendError(
                    "Pulp uploads of  {0} failed.".format(failed_uploads))
            if exceptions:
                raise CoprBackendError(f"Exceptions encountered: {exceptions}")

            return uploaded

    def create_repository_version(self, dirname, chroot, package_hrefs):
        """
        Create a new repository version by adding a list of RPMs to the latest repository version.
        """
        repository = self._get_repository(chroot, dirname)
        return self.client.add_content(repository, package_hrefs)

    def publish_repository(self, chroot, **kwargs):
        # Publishing occurs after each repository version is created.
        return True

    def delete_repository(self, chroot):
        repository = self._get_repository(chroot)
        distribution = self._get_distribution(chroot)
        self.client.delete_repository(repository)
        self.client.delete_distribution(distribution)

    def delete_project(self, dirname):
        prefix = "{0}/{1}".format(self.owner, dirname)
        response = self.client.list_distributions(prefix)
        distributions = response.json()["results"]
        for distribution in distributions:
            self.client.delete_distribution(distribution["pulp_href"])
            if distribution["repository"]:
                self.client.delete_repository(distribution["repository"])

    def delete_builds(self, dirname, chroot_builddirs, build_ids):
        # pylint: disable=too-many-locals
        result = True
        for chroot, subdirs in chroot_builddirs.items():
            # We don't upload results of source builds to Pulp
            if chroot == "srpm-builds":
                continue

            chroot_path = os.path.join(
                self.opts.destdir, self.owner, dirname, chroot)
            if not os.path.exists(chroot_path):
                self.log.error("%s chroot path doesn't exist", chroot_path)
                result = False
                continue

            repository = self._get_repository(chroot)
            # Find the RPMs by list of build ids
            content_response = self.client.get_content(build_ids, fields=["prn"])
            list_of_prns = [package["prn"] for package in content_response.json()["results"] ]
            self.client.delete_content(repository, list_of_prns)
            self.log.info("Deleted resources: %s", list_of_prns)

            # Delete results from the backend (logs, info files, etc)
            for subdir in subdirs:
                path = os.path.join(chroot_path, subdir)
                shutil.rmtree(path)
        return result

    def repository_exists(self, dirname, chroot, baseurl):
        repodata = "{0}/repodata/repomd.xml".format(baseurl)
        self.log.info("Checking that %s exists", repodata)
        try:
            SafeRequest().get(repodata)
        except RequestError:
            return False
        return True

    def fork_project(self, src_fullname, dst_fullname, builds_map):
        _dst_owner, dst_project = dst_fullname.split("/")
        for chroot, src_dst_dir in builds_map.items():
            if not chroot or not src_dst_dir:
                continue

            # It should be a dirname here but since forking CoprDirs is not
            # supported yet, we pass the project name
            # See https://github.com/fedora-copr/copr/issues/3820
            if not self.init_project(dst_project, chroot):
                self.log.error("Failed to init the dst project")
                return False

            for old_dir_name, new_dir_name in src_dst_dir.items():
                src_dir, dst_dir = old_dir_name, new_dir_name
                if not src_dir or not dst_dir:
                    continue

                src_build_id = int(src_dir.split("-")[0])
                dst_build_id = int(dst_dir.split("-")[0])

                # Forking doesn't support CoprDirs yet. For now, let's assume
                # all builds are in the main CoprDir.
                # See https://github.com/fedora-copr/copr/issues/3820
                dst_dirname = dst_fullname.split("/")[1]

                self._fork_build(
                    src_build_id,
                    dst_build_id,
                    src_fullname.split("/")[0],
                    src_fullname.split("/")[1],
                    dst_fullname.split("/")[0],
                    dst_fullname.split("/")[1],
                    dst_dirname,
                    chroot,
                )

            repository = self._get_repository(chroot)
            if not self.client.create_publication(repository):
                self.log.error("Failed to publish a repository")
                return False

        return True

    def _fork_build(self, src_build_id, dst_build_id, src_owner, src_project,
                    dst_owner, dst_project, dst_dirname, chroot):
        # pylint: disable=too-many-positional-arguments
        src_fullname = "{0}/{1}".format(src_owner, src_project)
        with TemporaryDirectory(prefix="copr-fork-") as tmp:
            response = self.client.get_content(
                [src_build_id],
                chroot,
                fields=["location_href"],
            )
            rpms = response.json()["results"]
            for rpm in rpms:
                filename = rpm["location_href"]
                url = "{0}/{1}/{2}/Packages/{3}/{4}".format(
                    self.opts.pulp_content_url,
                    src_fullname,
                    chroot,
                    filename[0],
                    filename,
                )

                self.log.info("Downloading %s", url)
                response = requests.get(url, timeout=60)

                # The package was likely in a CoprDir and forking supports only
                # main CoprDirs now. This check could be improved though.
                if response.status_code == 404:
                    self.log.error("Not found %s", url)
                    continue

                if not response.ok:
                    self.log.error("Failed to download %s because %s",
                                   url, response.reason)
                    return False

                path = os.path.join(tmp, os.path.basename(url))
                with open(path, "wb") as fp:
                    fp.write(response.content)

            # It is possible we didn't download any RPMs. That would be the case
            # for builds in CoprDirs.
            if not os.listdir(tmp):
                return None

            resign_rpms_in_dir(
                dst_owner,
                dst_project,
                tmp,
                chroot,
                self.opts,
                self.log,
            )
            rpms = self.find_build_results(tmp)
            result = self.upload_build_results(
                rpms,
                chroot,
                build_id=dst_build_id,
            )
            hrefs = [x["pulp_href"] for x in result.values()]

        self.log.info("Forked Pulp build %s as %s", src_build_id, dst_build_id)
        return self.create_repository_version(dst_dirname, chroot, hrefs)

    def _repository_name(self, chroot, dirname=None):
        return "/".join([
            self.owner,
            dirname or self.project,
            chroot,
        ])

    def _distribution_name(self, chroot, dirname=None, devel=None):
        # On backend we use /devel but in Pulp we cannot create subdirectories
        repository = self._repository_name(chroot, dirname)
        if devel is None:
            devel = self.devel
        if devel:
            return "{0}-devel".format(repository)
        return repository

    def _get_repository(self, chroot, dirname=None):
        name = self._repository_name(chroot, dirname)
        response = self.client.get_repository(name)
        return response.json()["results"][0]["pulp_href"]

    def _get_distribution(self, chroot, dirname=None, devel=None, full=False):
        if devel is None:
            devel = self.devel
        name = self._distribution_name(chroot, dirname, devel=devel)
        response = self.client.get_distribution(name)
        result = response.json()["results"][0]
        return result if full else result["pulp_href"]



from copr_common.worker_manager import QueueTask


class BackendQueueTask(QueueTask):
    """
    A base class for tasks processed by `BackendDispatcher` implementations
    """
    # pylint: disable=abstract-method

    @property
    def priority(self):
        return sum([self.frontend_priority, self.backend_priority])

    @property
    def frontend_priority(self):
        """
        A task priority is calculated based on a frontend priority preference
        and backend priority preference.

        This is the frontend priority number
        """
        return 0

    @property
    def backend_priority(self):
        """
        A task priority is calculated based on a frontend priority preference
        and backend priority preference.

        This is the backend priority number
        """
        return 0



"""
Message buses abstraction.
"""

import os
import logging
import copy
import json

try:
    from copr_messaging import schema
except ImportError:
    # copr_messaging is optional
    schema = None

from .constants import BuildStatus

try:
    import fedmsg
except ImportError:
    # fedmsg is optional
    fedmsg = None

try:
    import stomp
    from stomp import ConnectionListener as StompConnectionListener
except ImportError:
    # stomp is also optional
    stomp = None
    StompConnectionListener = object


def message_from_worker_job(style, topic, job, who, ip, pid):
    """
    Compat wrapper generating message object for messages defined before we
    switched to fedora-messaging jsonschema-aware model.  This can be removed
    once we can exepect that people are using copr-messaging module (and thus
    we can change message format without affecting the API).
    """
    if style == 'v1':
        content = {
            'user': job.submitter,
            'copr': job.project_name,
            'owner': job.project_owner,
            'pkg': job.package_name,
            'build': job.build_id,
            'chroot': job.chroot,
            'version': job.package_version,
            'status': job.status,
        }
        content.update({'ip': ip, 'who': who, 'pid': pid})

        message_types = {
            'build.start': {
                'what': "build start: user:{user} copr:{copr}" \
                        " pkg:{pkg} build:{build} ip:{ip} pid:{pid}",
                'class': schema.BuildChrootStartedV1,
            },
            'chroot.start': {
                'what': "chroot start: chroot:{chroot} user:{user}" \
                        " copr:{copr} pkg:{pkg} build:{build} ip:{ip} pid:{pid}",
                'class': schema.BuildChrootStartedV1DontUse,
            },
            'build.end': {
                'what': "build end: user:{user} copr:{copr} build:{build}" \
                        " pkg:{pkg} version:{version} ip:{ip} pid:{pid} status:{status}",
                'class': schema.BuildChrootEndedV1,
            },
        }

        content['what'] = message_types[topic]['what'].format(**content)
        message = message_types[topic]['class'](body=content)
        return message

    elif style == 'v1stomp':
        und = '(undefined)'
        content = {
            'user':        getattr(job, 'submitter', und),
            'copr':        getattr(job, 'project_name', und),
            'owner':       getattr(job, 'project_owner', und),
            'pkg':         getattr(job, 'package_name', und),
            'build':       getattr(job, 'build_id', und),
            'chroot':      getattr(job, 'chroot', und),
            'version':     getattr(job, 'package_version', und),
            'status':      getattr(job, 'status', und),
        }

        content['str_status'] = BuildStatus.string(content['status'])
        content.update({'ip': ip, 'who': who, 'pid': pid})

        msg_format = {
            'build.start': {
                'keys': {
                    'build': '{build}',
                    'owner': '{owner}',
                    'copr': '{copr}',
                    'submitter': '{user}',
                    'package': '{pkg}-{version}',
                    'chroot': '{chroot}',
                    'builder': '{ip}',
                    'status': '{str_status}',
                    'status_int': '{status}',
                },
                'class': schema.BuildChrootStartedV1Stomp,
            },
            'build.end': {
                'keys': {
                    'build': '{build}',
                    'owner': '{owner}',
                    'copr': '{copr}',
                    'submitter': '{user}',
                    'package': '{pkg}-{version}',
                    'chroot': '{chroot}',
                    'builder': '{ip}',
                    'status': '{str_status}',
                    'status_int': '{status}',
                },
                'class': schema.BuildChrootEndedV1Stomp,
            },
            'chroot.start': {
                'keys': {
                    'chroot': "{chroot}",
                },
                'class': schema.BuildChrootStartedV1StompDontUse,
            },
        }

        body = {}
        for key in msg_format[topic]['keys']:
            body[key] = msg_format[topic]['keys'][key].format(**content)

        return msg_format[topic]['class'](body=body)

    raise NotImplementedError


class MsgBus(object):
    """
    An "abstract" message bus class, don't instantiate!
    """
    messages = {}
    style = 'v1'
    bus_type = "msgbus"

    def __init__(self, opts, log=None):
        self.opts = opts
        # Fix bus_id soon enough.
        self.opts.bus_id = getattr(self.opts, 'bus_id', type(self).__name__)

        self.opts.bus_publish_retries = getattr(self.opts, 'bus_publish_retries', 5)

        self.log = log
        if not log:
            self.log = logging
            logging.basicConfig(level=logging.DEBUG)

        self.log.info("initializing bus")
        if hasattr(self.opts, 'messages'):
            self.messages.update(self.opts.messages)

    def _send_message(self, message):
        """
        Send message from fedora_messaging.message.Message (or subclass) object.
        The object is already validated.
        """
        raise NotImplementedError

    def send_message(self, message):
        """
        Validate and send message from fedora_messaging.message.Message (or
        subclass) object.
        """
        try:
            message.validate()
        except Exception: # pylint: disable=W0703
            self.log.exception("Failed to validate a message")
            return

        for attempt in range(1, self.opts.bus_publish_retries + 1):
            try:
                self._send_message(message)
                break
            except Exception: # pylint: disable=W0703
                # We don't want to halt the worker because of messaging.
                self.log.exception("Attempt %s to publish a message failed", attempt)

    def announce_job(self, msg_type, job, who, ip, pid):
        """
        Compat thing to be removed;  future types of messages (v2+) should be
        constructed at caller side, not in this class.
        """
        msg = message_from_worker_job(self.style, msg_type, job, who, ip, pid)
        self.send_message(msg)

    @property
    def info(self):
        """ string info about the bus """
        return "{} bus".format(self.bus_type)


class StompListener(StompConnectionListener):
    def __init__(self, msgbus):
        self.msgbus = msgbus

    def on_error(self, headers, message):
        self.msgbus.log.warning('received an error "%s"' % message)

    def on_disconnected(self):
        self.msgbus.log.warning('disconnected, trying to connect again..')
        self.msgbus.connect()


class MsgBusStomp(MsgBus):
    """
    Connect to STOMP bus and send messages.  Make sure you have correctly
    configured 'messages' attribute in every message bus configuration, no
    default messages here!
    """

    style = 'v1stomp'
    bus_type = "stomp"

    def connect(self):
        """
        connect (even repeatedly) to STOMP message bus
        """
        self.log.debug("connecting")
        self.conn.connect(
            # username/passcode can be None if ssl_key is used
            username=self.username,
            passcode=self.password,
            wait=True,
        )
        if not getattr(self.opts, 'destination', None):
            setattr(self.opts, 'destination', '/default')


    def __init__(self, opts, log=None):
        super(MsgBusStomp, self).__init__(opts, log)

        hosts = []
        # shortcuts
        if getattr(self.opts, 'hosts', None):
            assert type(self.opts.hosts) == list
            hosts += self.opts.hosts

        if getattr(self.opts, 'host', None):
            # TODO: Compat to be removed.
            self.log.warning("obsoleted 'host' parameter, use 'hosts' " \
                             "array (failover capable)")
            hosts.append((self.opts.host, self.opts.port))

        # Ensure integer ports.
        hosts = [(pair[0], int(pair[1])) for pair in hosts]

        self.conn = stomp.Connection(hosts)
        self.conn.set_listener('', StompListener(self))

        # allow dict.get() (with default None) method
        auth = {}
        if getattr(self.opts, 'auth', None):
            auth = self.opts.auth

        self.username = auth.get('username')
        self.password = auth.get('password')
        ssl_key  = auth.get('key_file')
        ssl_crt  = auth.get('cert_file')
        cacert = getattr(self.opts, 'cacert', None)

        if (ssl_key, ssl_crt, cacert) != (None, None, None):
            self.log.debug("ssl: key = %s, crt = %s, cacert = %s",
                ssl_key, ssl_crt, cacert)

            self.conn.set_ssl(
                for_hosts=hosts,
                key_file=ssl_key,
                cert_file=ssl_crt,
                ca_certs=cacert
            )

        self.connect()


    def _send_message(self, message):
        topic = message.topic
        body = message.body
        send_headers = {}
        send_headers['topic'] = topic

        destination = '{0}.{1}'.format(self.opts.destination, topic)

        # backward compat (before we mistakenly sent everything to the same
        # destination), drop once every consumer moved to copr-messaging module
        # and thus we can expect them to consume fixed messages too.
        if self.style == 'v1stomp':
            destination = self.opts.destination

        self.conn.send(body=json.dumps(body), headers=send_headers,
                       destination=destination,
                       content_type='application/json')


class MsgBusFedmsg(MsgBus):
    """
    Connect to fedmsg and send messages over it.
    """

    bus_type = "fedmsg"

    def __init__(self, log=None):
        # Hack to not require opts argument for now.
        opts = type('', (), {})
        opts.headers = {}

        super(MsgBusFedmsg, self).__init__(opts, log)

        fedmsg.init(name='relay_inbound', cert_prefix='copr', active=True)

    def _send_message(self, message):
        fedmsg.publish(modname='copr', topic=message.topic, msg=message.body)


class MsgBusFedoraMessaging(MsgBus):
    """
    Connect to fedora-messaging AMQP bus and send messages over it.
    """

    bus_type = "fedora-messaging"

    def __init__(self, opts, log=None):
        super(MsgBusFedoraMessaging, self).__init__(opts, log)
        # note this is not thread safe, only one bus of this type!
        os.environ['FEDORA_MESSAGING_CONF'] = opts.toml_config

    def _send_message(self, message):
        from fedora_messaging import api as fm_api, exceptions as fm_ex
        fm_api.publish(message)


class MessageSender:
    """
    Automatically send messages to all configured buses.
    """
    def __init__(self, backend_opts, name, log):
        self.log = log
        self.name = name

        msg_buses = []
        for bus_config in backend_opts.msg_buses:
            if bus_config.bus_type == 'stomp':
                msg_buses.append(MsgBusStomp(bus_config, log))
            elif bus_config.bus_type == 'fedora-messaging':
                msg_buses.append(MsgBusFedoraMessaging(bus_config, log))

        if backend_opts.fedmsg_enabled:
            msg_buses.append(MsgBusFedmsg(log))

        self.msg_buses = msg_buses
        self.pid = os.getpid()

    def announce(self, topic, job, host):
        """ Send message to all configured buses """
        for bus in self.msg_buses:
            self.log.info("Sending %s message in %s", bus.info, topic)
            bus.announce_job(
                topic, job,
                who=self.name,
                ip=host,
                pid=self.pid
            )



"""
Abstraction for RPM and SRPM builds on backend.
"""

from copr_common.worker_manager import (
    HashWorkerLimit,
    WorkerManager,
    PredicateWorkerLimit,
)
from copr_backend.worker_manager import BackendQueueTask
from copr_backend.helpers import get_chroot_arch


PRIORITY_SECTION_SIZE = 1000000

class BuildQueueTask(BackendQueueTask):
    """
    Build-task abstraction.  Needed for build our build scheduler (the
    WorkerManager class).

    Note that the worker counterpart (BackgroundWorker process) needs by far
    more information about the job to successfully process it.  But since we
    need to minimize the amount of informations downloaded by
    BuildDispatcher.load_jobs() method from frontent (performance reasons) we
    keep this in separate class.
    """
    def __init__(self, task):
        self._task = task
        self._backend_priority = 0
        try:
            int(self.id)
            self.source_build = True
        except ValueError:
            self.source_build = False

    @property
    def frontend_priority(self):
        """
        Set the "static" task priority, which can be determined purely from the
        Frontend task info.  We make sure that background builds are less
        prioritized (higher priority number) than non-background jobs, and
        source builds are more prioritized than RPM builds.  So we divide tasks
        to those categories:

        1. Source builds (priority > 0)
        2. Source background builds (priority > 1M)
        3. RPM builds (priority > 2M)
        4. RPM background builds (priority > 3M)
        """
        priority = self._task.get('priority', 0)

        if self.background:
            priority += 2*PRIORITY_SECTION_SIZE
        if not self.source_build:
            priority += PRIORITY_SECTION_SIZE
        return priority

    @property
    def background(self):
        """ True if this is "background" job (less priority) """
        return self._task.get("background", False)

    @property
    def backend_priority(self):
        return self._backend_priority

    @backend_priority.setter
    def backend_priority(self, value):
        self._backend_priority = value

    @property
    def id(self):
        return self._task['task_id']

    @property
    def build_id(self):
        """ Copr Frontend build.id this relates to. """
        return self._task['build_id']

    @property
    def chroot(self):
        """
        The chroot this task will be built in.  We return 'source' if this is
        source RPM build - in such case the build should be arch agnostic.
        """
        return self._task.get('chroot')

    @property
    def owner(self):
        """ Owner of the project this build belongs to """
        return self._task["project_owner"]

    @property
    def tags(self):
        """ Explicitly requested build tags """
        return self._task.get("tags", [])

    @property
    def requested_arch(self):
        """
        What is the requested "native" builder architecture for which this
        build task will be done.  We use this for limiting the build queue
        (i.e. separate limit for armhfp, even though such build process is
        emulated on x86_64).  Note that source builds also may require specific
        chroot (and thus architecture).
        """
        if not self.chroot:
            return None
        arch = get_chroot_arch(self.chroot)
        if arch.endswith("86"):
            # i386, i586, ...
            return "x86_64"
        return arch

    @property
    def sandbox(self):
        """
        Unique ID of "sandbox" to put the VM worker into.  Multiple builds can
        fall into the same sandbox, but only when it is absolutely safe (the
        same submitter, the same project, etc.).

        Frontend doesn't necessarily have to specify sandbox for each build,
        then we return None.  The consequence is that the allocated VM to such
        task is not possible to re-use for other purposes (before or after this
        task is processed).
        """
        return self._task.get('sandbox')


class ArchitectureWorkerLimit(PredicateWorkerLimit):
    """
    Limit the amount of concurrently running builds for the same architecture.
    """
    def __init__(self, architecture, limit):
        def predicate(x):
            return x.requested_arch == architecture
        super().__init__(predicate, limit, name="arch_{}".format(architecture))


class BlockedOwnersLimit(PredicateWorkerLimit):
    """
    Limit tasks from blocked owners to zero
    """
    def __init__(self, blocked_owners):
        def predicate(x):
            return x.owner in blocked_owners
        super().__init__(predicate, limit=0, name="blocked_users")


class ArchitectureUserWorkerLimit(HashWorkerLimit):
    """
    Limit number of machines of specific architecture we give to a single
    Copr owner (user or group).
    """
    def __init__(self, architecture, limit):
        def _hasher(job):
            if job.requested_arch != architecture:
                # When _hasher(job) returns None, no limit applies, see
                # HashWorkerLimit._groups.add() for more info.
                return None
            return f"{job.requested_arch}_{job.owner}"

        super().__init__(
            _hasher,
            limit,
            name=f"arch_{architecture}_owner",
        )


class UserSSHLimit(HashWorkerLimit):
    """
    Limit the number of builders that allow user SSH
    """
    def __init__(self, limit):
        def hasher(x):
            # We don't allow user SSH for SRPM builds, returning None will
            # make this unlimited
            if x.source_build:
                return None

            # Don't limit builds that doesn't allow user SSH
            # pylint: disable=protected-access
            if not x._task.get("allow_user_ssh"):
                return None
            return x.owner
        super().__init__(hasher, limit, name="userssh")


class BuildTagLimit(PredicateWorkerLimit):
    """
    Limit the amount of concurrently running builds per given build tag.
    """
    def __init__(self, tag, limit):
        def predicate(x):
            return tag in x.tags
        super().__init__(predicate, limit, name="tag_{}".format(tag))


class RPMBuildWorkerManager(WorkerManager):
    """
    Manager taking care of background build workers.
    """

    worker_prefix = 'rpm_build_worker'

    def start_task(self, worker_id, task):
        command = [
            "copr-backend-process-build",
            "--daemon",
            "--build-id", str(task.build_id),
            "--chroot", "srpm-builds" if task.source_build else task.chroot,
            "--worker-id", worker_id,
        ]
        self.log.info("running worker: %s", " ".join(command))
        self.start_daemon_on_background(command)

    def finish_task(self, worker_id, task_info):
        self.get_task_id_from_worker_id(worker_id)
        return True
